###############################################################################
### This is a reimplementation of workspaceCombiner library in python
### original author: Hongtao Yang
###############################################################################
import os
import re
import copy
import json
import time
import yaml
from typing import Optional, Union, List, Dict

import numpy as np

import ROOT

import quickstats
from quickstats import semistaticmethod, Timer, AbstractObject, GeneralEnum
from quickstats.components import ExtendedModel
from quickstats.utils.xml_tools import TXMLTree
from quickstats.utils.common_utils import format_delimiter_enclosed_text, remove_list_duplicates
from quickstats.maths.numerics import is_float, pretty_value
from quickstats.components.basics import WSArgument
from quickstats.components.workspaces import XMLWSBase, WSObjectType, AsimovHandler

class ModifierAction(GeneralEnum):
    NORMAL     = 0
    CONSTRAINT = 1
    
class XMLWSModifier(XMLWSBase):
    
    def __init__(self, source:Union[str, Dict], basedir:Optional[str]=None,
                 minimizer_config:Optional[Dict]=None,
                 unlimited_stack:bool=True,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(source=source, basedir=basedir,
                         unlimited_stack=unlimited_stack,
                         verbosity=verbosity)
        self.minimizer_config = minimizer_config
        self.load_extension()
        self.initialize(source=source)
        
    def initialize(self, source:Union[str, Dict]):
        self.config = None
        self.actions = {
            'map': [],
            'define': [],
            'redefine': [],
            'asimov': [],
            'constraint': [],
            'add_product_terms': {},
            'add_syst': []
        }
        self.actions['rename'] = {
            'workspace'   : {},
            'model_config': {},
            'dataset'     : {},
            'variable'    : {}
        }
        if isinstance(source, str):
            ext = os.path.splitext(source)[-1]
            if ext == ".xml":
                self.parse_config_xml(source)
            elif ext == ".json":
                self.parse_config_json(source)
            elif ext == ".yaml":
                self.parse_config_yaml(source)
            else:
                raise ValueError(f"unsupported file format: {ext}")
        elif isinstance(source, dict):
            self.parse_config_dict(source)
        else:
            raise ValueError(f"invalid input format: {source}")
        if len(self.actions['add_product_terms']) > 0:
            if (quickstats.root_version < (6, 26, 0)):
                raise RuntimeError("\"add product terms\" is only supported after ROOT 6.26")
            
    def parse_config_dict(self, source:Dict):
        source = copy.deepcopy(source)
        _actions = source.pop("actions", {})
        config = {
            "input_file"         : source.pop("input_file", "dummy"),
            "output_file"        : source.pop("output_file", "dummy"),
            "model_name"         : source.pop("model_name", None),
            "workspace_name"     : source.pop("workspace_name", None),
            "model_config_name"  : source.pop("model_config_name", None),
            "data_name"          : source.pop("data_name", "combData"),
            "poi_names"          : source.pop("poi_names", None),
            "snapshot_nuis"      : source.pop("snapshot_nuis", []),
            "snapshot_globs"     : source.pop("snapshot_globs", []),
            "snapshot_pois"      : source.pop("snapshot_pois", []),
            "snapshot_all"       : source.pop("snapshot_all", []),
            "snapshot_list"      : source.pop("snapshot_list", []),
            "dataset_list"       : source.pop("dataset_list", None),
            "fix_parameters"     : source.pop("fix_parameters", None),
            "profile_parameters" : source.pop("profile_parameters", None),
            "strict"             : source.pop("strict", True)
        }
        if source:
            unknown_attributes = list(source)
            self.stdout.warning(f"WARNING: The following unrecognized configuration attribute(s) will be ignored: "
                                f"{', '.join(unknown_attributes)}")
        actions = {
            "map"        : _actions.pop("map", []),
            "define"     : _actions.pop("define", []),
            "redefine"   : _actions.pop("redefine", []),
            "asimov"     : _actions.pop("asimov", []),
            "constraint" : [],
            "add_product_terms": _actions.pop("add_product_terms", {}),
            'add_syst': _actions.pop("add_syst", [])
        }
        
        # handle constraint definition
        constraints = _actions.pop("constraint", [])
        for constraint_dict in constraints:
            constraint_node = {
                'attrib': constraint_dict
            }
            expr = self._get_node_attrib(constraint_node, "Name", required=True)
            constr_expr = self.parse_constr_expression(expr, constraint_node)
            if constr_expr['file'] is None:
                actions['define'].append(expr)
            actions['constraint'].append(constr_expr)

        _rename = _actions.pop("rename", {})
        rename = {
            "workspace"    : _rename.pop("workspace", {}),
            "model_config" : _rename.pop("model_config", {}),
            "dataset"      : _rename.pop("dataset", {}),
            "variable"     : _rename.pop("variable", {})
        }
        actions["rename"] = rename
        if _actions:
            unknown_attributes = list(_actions)
            self.stdout.warning(f"WARNING: Unknown action type: {', '.join(unknown_attributes)}. "
                                "Ignoring...")
        if _rename:
            unknown_attributes = list(_rename)
            self.stdout.warning(f"WARNING: Unknown object type in rename: {', '.join(unknown_attributes)}. "
                                "Ignoring...")
        
        self.config = config
        self.actions = actions
    
    def parse_config_json(self, filename:str):
        with open(filename, 'r') as f:
            source = json.load(f)
        self.parse_config_dict(source)
        
    def parse_config_yaml(self, filename:str):
        with open(filename, 'r') as f:
            source = yaml.safe_load(f)
        self.parse_config_dict(source)        
    
    def parse_config_xml(self, filename:str):
        root = TXMLTree.load_as_dict(filename)
        config = {
            'input_file'       : self._get_node_attrib(root, "InFile", required=True),
            'output_file'      : self._get_node_attrib(root, "OutFile", required=True),
            'model_name'       : self._get_node_attrib(root, "ModelName", required=False),
            'workspace_name'   : self._get_node_attrib(root, "WorkspaceName", required=False),
            'model_config_name': self._get_node_attrib(root, "ModelConfigName", required=False),
            'data_name'        : self._get_node_attrib(root, "DataName", required=False, default="combData"),
            'poi_names'        : self._get_node_attrib(root, "POINames", required=False, dtype="str_list"),
            'snapshot_nuis'    : self._get_node_attrib(root, "SnapshotNP", required=False, default=[], dtype="str_list"),
            'snapshot_globs'   : self._get_node_attrib(root, "SnapshotGO", required=False, default=[], dtype="str_list"),
            'snapshot_pois'    : self._get_node_attrib(root, "SnapshotPOI", required=False, default=[], dtype="str_list"),
            'snapshot_all'     : self._get_node_attrib(root, "SnapshotAll", required=False, default=[], dtype="str_list"),
            'snapshot_list'    : self._get_node_attrib(root, "SnapshotList", required=False, default=[], dtype="str_list"),
            'dataset_list'     : self._get_node_attrib(root, "DatasetList", required=False, default=None, dtype="str_list"),
            "fix_parameters"     : self._get_node_attrib(root, "FixParameters", required=False, default=None),
            "profile_parameters" : self._get_node_attrib(root, "ProfileParameters", required=False, default=None),
            'strict'           : self._get_node_attrib(root, "Strict", required=False, default="true", dtype="bool")
        }
        # for compatibility with older version of workspaceCombiner
        if config['workspace_name'] == "dummy":
            config['workspace_name'] = None
        if config['model_config_name'] == "dummy":
            config['model_config_name'] = None
        self.config = config
        
        # parse child nodes
        nodes = root['children']
        for node in nodes:
            self.parse_action_node(node)
            
    def parse_action_node(self, node:Dict):
        tag = node['tag']
        if tag == "Item":
            expr        = self._get_node_attrib(node, "Name", required=True)
            action_type = self._get_node_attrib(node, "Type", required=False,
                                                default=ModifierAction.NORMAL)
            action_type = ModifierAction.parse(action_type)
            if action_type == ModifierAction.CONSTRAINT:
                constr_expr = self.parse_constr_expression(expr, node)
                # define pdf that is not imported externally
                if constr_expr['file'] is None:
                    self.actions['define'].append(expr)
                self.actions['constraint'].append(constr_expr)
            else:
                self.actions['define'].append(expr)
        elif tag == "Map":
            expr = self._get_node_attrib(node, "Name", required=True)
            self.actions['map'].append(expr)
        elif tag == "Asimov":
            definitions = node['attrib']
            self.actions['asimov'].append(definitions)
        elif tag == "Rename":
            subnodes = node['child']
            for subnode in subnodes:
                self.parse_rename_node(subnode)
        elif tag == "AddProductTerm":
            name = self._get_node_attrib(node, "Name", required=True)
            extra_terms = self._get_node_attrib(node, "Terms", required=True, dtype="str_list")
            if name in self.actions['add_product_terms']:
                raise ValueError("AddProductTerm action defined multiple times for the variable "
                                 f"\"{name}\"")
            self.actions['add_product_terms'][name] = extra_terms
        elif tag == "AddNPSet":
            nuis_names = self._get_node_attrib(node, "Name", required=True, dtype="str_list")
            self.actions['add_syst'] += nuis_names
        else:
            raise RuntimeError(f"unknown item `{tag}`")
    
    def parse_rename_node(self, node:Dict):
        tag = node['tag']
        new_name = self._get_node_attrib(node, "New", required=True)
        if tag == 'Workspace':
            old_name_required = False
            target = 'workspace'
        elif tag == 'ModelConfig':
            old_name_required = False
            target = 'model_config'
        elif tag == 'Dataset':
            old_name_required = True
            target = 'dataset'
        elif tag == 'Variable':
            old_name_required = True
            target = 'variable'
        else:
            raise RuntimeError(f"unknown item `{tag}`")
        old_name = self._get_node_attrib(node, "Old", required=old_name_required)
        if old_name in self.actions['rename'][target]:
            raise RuntimeError(f"the {target.replace('_','')} \"{old_name}\" is renamed more than once")
        self.actions['rename'][target][old_name] = new_name
    
    def parse_constr_expression(self, expr:str, node:Dict):
        pdf_name, _ = self._get_object_name_and_type_from_expr(expr)
        nuis_name = self._get_node_attrib(node, "NP", required=False, default=[], dtype="str_list")
        glob_name = self._get_node_attrib(node, "GO", required=False, default=[], dtype="str_list")
        filename  = self._get_node_attrib(node, "FileName", required=False)
        independent  = self._get_node_attrib(node, "Independent", required=False, default="false", dtype="bool")
        result = {
            'pdf': pdf_name,
            'nuis': nuis_name,
            'glob': glob_name,
            'file': filename,
            'independent': independent
        }
        return result
    
    def sanity_check(self):
        if self.config is None:
            raise RuntimeError("core configuration not set")
        if len(self.actions['rename']['workspace']) > 1:
            raise RuntimeError("workspace is renamed more than once")
        if len(self.actions['rename']['model_config']) > 1:
            raise RuntimeError("model config is renamed more than once")
    
    def create_modified_workspace(self, infile:Optional[str]=None, outfile:Optional[str]=None,
                                  import_class_code:bool=True, recreate:bool=True):
        with Timer() as t:
            self.sanity_check()
            # override path to the input workspace
            if infile is None:
                infile = self.config['input_file']
            # override path to the output workspace
            if outfile is None:
                outfile = self.config['output_file']
            ws_name  = self.config['workspace_name']
            mc_name  = self.config['model_config_name']
            self.stdout.info(f"INFO: Begin modification of the workspace \"{infile}\".")
            model    = ExtendedModel(infile, ws_name=ws_name, mc_name=mc_name,
                                     data_name=None, verbosity="WARNING")
            ws_orig  = model.workspace
            mc_orig  = model.model_config
            ws_name  = ws_orig.GetName()
            mc_name  = mc_orig.GetName()

            # create temporary workspace
            ws_tmp = ROOT.RooWorkspace(ws_name)

            title_str = format_delimiter_enclosed_text("Step 1: Redefine objects", "-")
            self.stdout.info(title_str)
            flag = self.redefine_objects(ws_orig, ws_tmp)
            # nothing is done
            if not flag:
                self.stdout.info("INFO: No objects are redefined.")

            title_str = format_delimiter_enclosed_text("Step 2: Create new objects", "-")
            self.stdout.info(title_str)
            flag = self.implement_external_pdfs(ws_tmp)
            flag |= self.implement_objects(ws_tmp)
            # nothing is done
            if not flag:
                self.stdout.info("INFO: No new objects are defined.")
            
            dataset_list = self.config['dataset_list']
            # need to import datasets first so the variables are also imported into the workspace
            self.import_datasets(ws_orig, ws_tmp, dataset_list)

            title_str = format_delimiter_enclosed_text("Step 3: Rename variables", "-")
            self.stdout.info(title_str)
            rename_map = self.get_rename_map(ws_orig, ws_tmp)
            actual_rename_map = self.get_actual_rename_map(rename_map, ws_orig, ws_tmp)
            self.rename_variables(mc_orig, ws_tmp, rename_map)
            if not rename_map:
                self.stdout.info("INFO: No variables are renamed.")

            title_str = format_delimiter_enclosed_text("Step 4: Making output workspace", "-")
            self.stdout.info(title_str)

            # create final workspace
            if len(self.actions['constraint']) > 0:
                ws_final = ROOT.RooWorkspace(ws_name)
                sim_pdf = self.remake_simultaneous_pdf(mc_orig, ws_tmp)
                self.import_object(ws_final, sim_pdf)
                self.import_datasets(ws_orig, ws_final, dataset_list)
            else:
                ws_final = ws_tmp        

            # post processing pdfs, functions and variables
            self.post_process_workspace(ws_final)

            mc_final = self.create_model_config(mc_orig, ws_final, actual_rename_map)

            self.rename_objects(ws_final, mc_final)

            self.import_object(ws_final, mc_final, silent=False)
            if import_class_code:
                self.import_class_code(ws_final)

            self.create_snapshots(ws_orig, mc_orig, ws_final, mc_final, rename_map)
            
            self.create_generic_objects(ws_orig, ws_final, rename_map)

            self.generate_asimov(ws_final)

            self.setup_parameters(ws_final)

            outdir  = os.path.dirname(outfile)
            if outdir and (not os.path.exists(outdir)):
                os.makedirs(outdir)

            ws_final.writeToFile(outfile, recreate)
            self.stdout.info(f'INFO: Saved output workspace as "{outfile}".')
        self.stdout.info(f"INFO: Total time taken: {t.interval:.3f}s")
    
    def post_process_workspace(self, ws:ROOT.RooWorkspace):
        self.add_product_terms(ws)
    
    def add_product_terms(self, ws:ROOT.RooWorkspace):
        for target_name, extra_terms in self.actions["add_product_terms"].items():
            prod_var = self.get_workspace_arg(ws, target_name, strict=True)
            if not isinstance(prod_var, ROOT.RooProduct):
                class_name = prod_var.ClassName()
                raise RuntimeError(f"failed to add product term to the variable \"{target_name}\": "
                                   f"expect a RooProduct instance but {class_name} received")
            for extra_term in extra_terms:
                member_var = self.get_workspace_arg(ws, extra_term, strict=True)
                prod_var.addTerm(member_var)
            self.stdout.info(f"INFO: The following product terms are added to the variable {target_name}: "
                             f"{', '.join(extra_terms)}")
    
    def setup_parameters(self, ws:ROOT.RooWorkspace):
        data_name = self.config['data_name']
        model = ExtendedModel(ws, data_name=data_name, verbosity="WARNING")
        model.stdout.verbosity = self.stdout.verbosity
        if self.config["fix_parameters"] is not None:
            model.fix_parameters(self.config["fix_parameters"])
        if self.config["profile_parameters"] is not None:
            model.profile_parameters(self.config["profile_parameters"])

    def generate_asimov(self, ws:ROOT.RooWorkspace):
        n_asimov = len(self.actions['asimov'])
        if n_asimov == 0:
            return None
        data_name = self.config['data_name']
        asimov_handler = AsimovHandler(ws, data_name=data_name,
                                       minimizer_config=self.minimizer_config)
        for attributes in self.actions['asimov']:
            asimov_handler.generate_single_asimov(attributes)
            
    def copy_variable(self, old_variable:ROOT.RooRealVar, new_variable:ROOT.RooRealVar):
        copied_variable = new_variable.Clone()
        copied_variable.setVal(old_variable.getVal())
        copied_variable.setRange(old_variable.getMin(), old_variable.getMax())
        copied_variable.setConstant(old_variable.isConstant())
        copied_variable.setError(old_variable.getError())
        return copied_variable
            
    def copy_variables(self, variables:ROOT.RooArgSet, new_ws:ROOT.RooWorkspace, rename_map:Optional[Dict]=None):
        if rename_map is None:
            rename_map = {}
        renamed_variables_old = ROOT.RooArgSet()
        renamed_variables_new = ROOT.RooArgSet()
        for variable in variables:
            name = variable.GetName()
            if name in rename_map:
                new_name = rename_map[name]
                # note only renaming of RooRealVar instance will be considered
                new_variable = new_ws.var(new_name)
                if not new_variable:
                    self.stdout.warning(f'WARNING: The argset variable "{name}" not found in the new workspace. Skipped.')
                # avoid modifying the original value
                copied_variable = self.copy_variable(variable, new_variable)
                renamed_variables_old.add(variable)
                renamed_variables_new.add(copied_variable)
        variables.remove(renamed_variables_old)
        variables.add(renamed_variables_new)
                               
    def copy_snapshot_variables(self, old_ws:ROOT.RooWorkspace, old_mc:ROOT.RooStats.ModelConfig,
                                new_ws:ROOT.RooWorkspace, new_mc:ROOT.RooStats.ModelConfig,
                                snapshot_name:str, variable_type:Optional[Union[WSArgument, str]]=None,
                                rename_map:Optional[Dict]=None):
        snapshot = old_ws.getSnapshot(snapshot_name)
        if not snapshot:
            self.stdout.warning(f"WARNING: Snapshot {snapshot_name} not found in the original workspace")
        # get variables from orginal workspace, make sure to make a clone
        if variable_type is None:
            variables = snapshot.Clone()
        else:
            target_variables = self.get_variables(old_mc, variable_type)
            variables = snapshot.Clone().selectCommon(target_variables)
        if rename_map is None:
            rename_map = {}
        self.stdout.info(f"INFO: Copying snapshot \"{snapshot_name}\" from the original workspace "
                         "to the new workspace")
        self.copy_variables(variables, new_ws, rename_map)
        # important: the third argument decide whether to take values from the argset or from the workspace
        new_ws.saveSnapshot(snapshot_name, variables, True)
        
    def get_variables(self, mc:ROOT.RooStats.ModelConfig, variable_type:Union[WSArgument, str]):
        _variable_type = WSArgument.parse(variable_type)
        if _variable_type == WSArgument.NUISANCE_PARAMETER:
            variables = mc.GetNuisanceParameters()
        elif _variable_type == WSArgument.GLOBAL_OBSERVABLE:
            variables = mc.GetGlobalObservables()
        elif _variable_type == WSArgument.POI:
            variables = mc.GetParametersOfInterest()
        elif _variable_type == WSArgument.CORE:
            variables = ROOT.RooArgSet()
            variables.add(mc.GetNuisanceParameters())
            variables.add(mc.GetGlobalObservables())
            variables.add(mc.GetParametersOfInterest())
        else:
            raise RuntimeError(f"unsupported variable type: {variable_type}")
        return variables

    def create_snapshots(self, old_ws:ROOT.RooWorkspace, old_mc:ROOT.RooStats.ModelConfig,
                         new_ws:ROOT.RooWorkspace, new_mc:ROOT.RooStats.ModelConfig,
                         rename_map:Optional[Dict]=None):
        # nuisance parameter snapshots
        nuis_snapshot_names = self.config['snapshot_nuis']
        for nuis_snapshot_name in nuis_snapshot_names:
            self.copy_snapshot_variables(old_ws, old_mc, new_ws, new_mc, nuis_snapshot_name,
                                         WSArgument.NUISANCE_PARAMETER, rename_map)
        # global observable snapshots                  
        globs_snapshot_names = self.config['snapshot_globs']
        for globs_snapshot_name in globs_snapshot_names:
            self.copy_snapshot_variables(old_ws, old_mc, new_ws, new_mc, globs_snapshot_name,
                                         WSArgument.GLOBAL_OBSERVABLE, rename_map)
        # poi snapshots 
        pois_snapshot_names = self.config['snapshot_pois']
        for pois_snapshot_name in pois_snapshot_names:
            self.copy_snapshot_variables(old_ws, old_mc, new_ws, new_mc, pois_snapshot_name,
                                         WSArgument.POI, rename_map)
        # nuisance parameter + global observable + poi snapshots 
        vars_snapshot_names = self.config['snapshot_all']
        for vars_snapshot_name in vars_snapshot_names:
            self.copy_snapshot_variables(old_ws, old_mc, new_ws, new_mc, vars_snapshot_name,
                                         WSArgument.CORE, rename_map)
        # any variables
        vars_snapshot_names = self.config['snapshot_list']
        for vars_snapshot_name in vars_snapshot_names:
            self.copy_snapshot_variables(old_ws, old_mc, new_ws, new_mc, vars_snapshot_name,
                                         None, rename_map)
            
    def create_generic_objects(self, old_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace,
                               rename_map:Optional[Dict]=None):
        generic_objects = old_ws.allGenericObjects()
        # only create objects of type RooArgSet
        generic_objects = [obj for obj in generic_objects if isinstance(obj, ROOT.RooArgSet)]
        if rename_map is None:
            rename_map = {}
        for generic_object in generic_objects:
            name = generic_object.GetName()
            if not name:
                self.stdout.warning('WARNING: Found generic object without name. '
                                    'The object will not be imported to the new workspace')
                continue
            self.stdout.info(f'INFO: Copying generic object "{name}" from the original workspace to the new workspace"')
            if all(isinstance(obj, ROOT.RooRealVar) for obj in generic_object):
                argset = generic_object.Clone()
                self.copy_variables(argset, new_ws, rename_map)
            else:
                argset = self.recreate_argset(generic_object, new_ws, rename_map)
            new_ws.Import(argset, name, True)
            
            
    def recreate_argset(self, argset:ROOT.RooArgSet, ws:ROOT.RooWorkspace, rename_map:Optional[Dict]=None):
        argset_name = argset.GetName()
        new_argset = ROOT.RooArgSet()
        new_argset.setName(argset_name)
        for arg in argset:
            name = arg.GetName()
            name = rename_map.get(name, name)
            new_arg = ws.obj(name)
            if not new_arg:
                raise RuntimeError(f'WARNING: The variable "{name}" from the argset "{argset_name}" '
                                   f'not found in the new workspace.')
            if isinstance(new_arg, ROOT.RooRealVar):
                copied_arg = self.copy_variable(arg, new_arg)
                new_argset.add(copied_arg)
            else:
                new_argset.add(new_arg)
        return new_argset
 
    def create_model_config(self, old_mc:ROOT.RooStats.ModelConfig, new_ws:ROOT.RooWorkspace,
                            rename_map:Optional[Dict]=None):
        if rename_map is None:
            rename_map = {}
            
        strict = self.config['strict']
        
        # start preparing model config
        mc_name = old_mc.GetName()
        mc = ROOT.RooStats.ModelConfig(mc_name, new_ws)

        # set pdf
        pdf_name = old_mc.GetPdf().GetName()
        pdf = new_ws.pdf(pdf_name)        
        mc.SetPdf(pdf)
        
        # set POIs
        self.stdout.debug("List of POIs in the new parameterization:")
        pois = ROOT.RooArgSet()
        poi_names = self.config['poi_names']
        # use old poi names if not specified
        if poi_names is None:
            poi_names = [i.GetName() for i in old_mc.GetParametersOfInterest()]
            poi_names = [rename_map.get(poi_name, poi_name) for poi_name in poi_names]
        for poi_name in poi_names:
            poi = self._get_relevant_variable(poi_name, new_ws, pdf)
            if not poi:
                if strict:
                    raise RuntimeError(f"POI {poi_name} does not exist")
                else:
                    self.stdout.warning(f"WARNING: POI {poi_name} does not exist. Skipping.")
                    continue
            pois.add(poi)
            self.stdout.info(f"INFO: Added POI \"{poi_name}\".")
        mc.SetParametersOfInterest(pois)
        
        # nuisance parameters and global observables from original workspace
        old_nuis = old_mc.GetNuisanceParameters()
        if old_nuis:
            nuis_names = [nuis.GetName() for nuis in old_nuis]
            nuis_names = [rename_map.get(nuis, nuis) for nuis in nuis_names]
        else:
            nuis_names = []
        old_globs = old_mc.GetGlobalObservables()
        if old_globs:
            glob_names = [glob.GetName() for glob in old_globs]
            glob_names = [rename_map.get(glob, glob) for glob in glob_names]
        else:
            glob_names = []
            
        # newly defined nuisance_parameters and global observables
        additional_constraints = self.actions['constraint']
        new_nuis = []
        for item in additional_constraints:
            new_nuis   += item['nuis']
            nuis_names += item['nuis']
            glob_names += item['glob']
            
        # additional (unconstrained) nuisance parameters
        extra_nuis = self.actions['add_syst']
        nuis_names += extra_nuis
        
        if len(new_nuis) > 0:
            self.stdout.info("INFO: The following newly defined nuisance parameter(s) will be added to ModelConfig")
            for nuis_name in new_nuis:
                self.stdout.info(f"    {nuis_name}")
        if len(extra_nuis) > 0:
            self.stdout.info("INFO: The following additional nuisance parameter(s) will be added to ModelConfig")
            for nuis_name in extra_nuis:
                self.stdout.info(f"    {nuis_name}")
        
        nuis_names = remove_list_duplicates(nuis_names)
        # set nuisance parameters
        nuisance_parameters = ROOT.RooArgSet()
        for nuis_name in nuis_names:
            nuis = self._get_relevant_variable(nuis_name, new_ws, pdf)
            if not nuis:
                self.stdout.warning(f"WARNING: The nuisance parameter {nuis_name} no longer exists in "
                                    "the new workspace. It will be removed from the new ModelConfig.")
                continue
            if nuis.isConstant():
                self.stdout.warning(f"WARNING: The nuisace parameter {nuis_name} is initialized as a constant. "
                                    "It will be floated in the new workspace.")
                nuis.setConstant(False)
            nuisance_parameters.add(nuis)
                
        # set global observables
        global_observables = ROOT.RooArgSet()
        for glob_name in glob_names:
            glob = self._get_relevant_variable(glob_name, new_ws, pdf)
            if not glob:
                self.stdout.warning(f"WARNING: The global observable {glob_name} no longer exists in "
                                    "the new workspace. It will be removed from the new ModelConfig.")
                continue
            if not glob.isConstant():
                self.stdout.warning(f"WARNING: The global observable {glob_name} is initialized as a floating "
                                    "parameter. It will be set to constant in the new workspace.")

                glob.setConstant(True)
            global_observables.add(glob)

        # set observables
        observables = ROOT.RooArgSet()
        old_obs = old_mc.GetObservables()
        if old_obs:
            for obs in old_obs:
                obs_name = obs.GetName()
                new_obs = new_ws.obj(obs_name)
                if not new_obs:
                    raise RuntimeError(f"observable {obs_name} no longer exists in the new workspace")
                observables.add(new_obs)
        mc.SetNuisanceParameters(nuisance_parameters)
        mc.SetGlobalObservables(global_observables)
        mc.SetObservables(observables)
        
        return mc
    
    @semistaticmethod
    def extract_ws_variables(self, ws:ROOT.RooWorkspace, variables:ROOT.RooArgSet, strict:bool=True):
        extracted_variables = ROOT.RooArgSet()
        for variable in variables:
            variable_name = variable.GetName()
            extracted_variable = ws.arg(variable_name)
            if not extracted_variable:
                if strict:
                    raise RuntimeError(f"missing variable {variable_name} in the workspace {ws.GetName()}")
                else:
                    self.stdout.warning(f"WARNING: No variable {variable_name} found in the workspace {ws.GetName()}")
                    continue
            extracted_variables.add(extracted_variable, True)
        return extracted_variables

    def rename_variables(self, old_mc:ROOT.RooStats.ModelConfig, new_ws:ROOT.RooWorkspace, rename_map:Dict):
        old_var_expr = ",".join(list(rename_map.keys()))
        new_var_expr = ",".join(list(rename_map.values()))
        getattr(new_ws, "import")(old_mc.GetPdf(),
                                  ROOT.RooFit.RenameVariable(old_var_expr, new_var_expr),
                                  ROOT.RooFit.RecycleConflictNodes())
        if rename_map:
            self.stdout.info("INFO: Renamed variables\n")
            rename_table = "\n".join(self.get_name_mapping_str_arrays(rename_map))
            self.stdout.info(rename_table)
        
    def rename_objects(self, ws:ROOT.RooWorkspace, mc:ROOT.RooStats.ModelConfig):
        if len(self.actions['rename']['workspace']) > 0:
            self.stdout.info("INFO: Renamed workspace\n")
            rename_map = {}
            for old_name, new_name in self.actions['rename']['workspace'].items():
                rename_map[ws.GetName()] = new_name
                ws.SetName(new_name)
            rename_table = "\n".join(self.get_name_mapping_str_arrays(rename_map))
            self.stdout.info(rename_table)
        if len(self.actions['rename']['model_config']) > 0:
            self.stdout.info("INFO: Renamed model config\n")
            rename_map = {}
            for old_name, new_name in self.actions['rename']['model_config'].items():
                rename_map[mc.GetName()] = new_name
                mc.SetName(new_name)
            rename_table = "\n".join(self.get_name_mapping_str_arrays(rename_map))
            self.stdout.info(rename_table)
        if len(self.actions['rename']['dataset']) > 0:
            rename_map = {}            
            for old_name, new_name in self.actions['rename']['dataset'].items():
                if old_name == new_name:
                    continue
                dataset = ws.data(old_name)
                if not dataset:
                    raise RuntimeError(f"dataset \"{old_name}\" not found in the original workspace")
                check_dataset = ws.data(new_name)
                if check_dataset:
                    raise RuntimeError(f"cannot rename dataset from \"{old_name}\" to \"{new_name}\": "
                                       f"the dataset \"{new_name}\" already exists in the original workspace")
                rename_map[old_name] = new_name
                dataset.SetName(new_name)
            if rename_map:
                self.stdout.info("INFO: Renamed dataset\n")
                rename_table = "\n".join(self.get_name_mapping_str_arrays(rename_map))
                self.stdout.info(rename_table)                
        
    def import_datasets(self, old_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace,
                        dataset_names:Optional[List[str]]=None):
        if dataset_names is None:
            datasets = old_ws.allData()
        else:
            datasets = []
            for dataset_name in dataset_names:
                dataset = old_ws.data(dataset_name)
                if not dataset:
                    raise RuntimeError(f"dataset {dataset_name} does not exist in the original workspace")
                datasets.append(dataset)
        for dataset in datasets:
            getattr(new_ws, "import")(dataset)
            self.stdout.info(f"INFO: Imported dataset \"{dataset.GetName()}\".")
            
    def redefine_objects(self, old_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace):
        flag = False
        for i, expr in enumerate(self.actions['redefine']):
            self.stdout.info(f"INFO: (Item {i}) {expr}")
            obj_name, obj_type = self._get_object_name_and_type_from_expr(expr)
            arg = old_ws.arg(obj_name)
            if not arg:
                raise RuntimeError(f"object {obj_name} does not exist in the original workspace")
            if obj_type == WSObjectType.VARIABLE:
                result = re.search(r"\[(.+)\]", expr)
                if not result:
                    raise RuntimeError(f"invalid variable expression {expr}")
                self.import_expression(new_ws, expr)
                new_var = new_ws.var(obj_name)
                tokens = result.group(1).split(",")
                # only modify the value
                if len(tokens) == 1:
                    old_range = arg.getRange()
                    # need to restore the range
                    new_var.setRange(old_range[0], old_range[1])
                # only modify the range
                elif len(tokens) == 2:
                    # need to restore the value
                    new_var.setVal(arg.getVal())
                # need to restore constant state
                new_var.setConstant(arg.isConstant())
            else:
                self.import_expression(new_ws, expr)
            flag = True
        return flag
        
    def implement_objects(self, ws:ROOT.RooWorkspace):
        flag = False
        for i, expr in enumerate(self.actions['define']):
            self.stdout.info(f"INFO: (Item {i}) {expr}")
            if "FlexibleInterpVar" in expr:
                self.implement_flexible_interp_var(ws, expr)
            elif "RooMultiVarGaussian" in expr:
                self.implement_multi_var_gaussian(ws, expr)
            else:
                self.import_expression(ws, expr)
            flag = True
        return flag
                
    def implement_external_pdfs(self, ws:ROOT.RooWorkspace):
        flag = False
        for i, item in enumerate(self.actions['constraint']):
            pdf_name = item['pdf']
            file = item['file']
            if file is not None:
                self.load_external_pdf(file, pdf_name, ws)
                self.stdout.info(f"INFO: Loaded external pdf {pdf_name} from {file}")
                flag = True
        return flag
                
    def load_external_pdf(self, ext_rfile:str, pdf_name:str, ws:ROOT.RooWorkspace):
        model = ExtendedModel(ext_rfile, data_name=None, verbosity="WARNING")
        pdf = model.workspace.pdf(pdf_name)
        if not pdf:
            raise RuntimeError(f"pdf {pdf_names} not found in the workspace loaded from {ext_rfile}")
        self.import_object(ws, pdf)
    
    def get_actual_rename_map(self, rename_map:Dict, old_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace):
        actual_rename_map = {}
        for old_name, new_name in rename_map.items():
            # if the object with the new name isalready defined, it's not a renaming
            if (old_ws.arg(new_name)) or (new_ws.arg(new_name)):
                continue
            actual_rename_map[old_name] = new_name
        return actual_rename_map
    
    def get_rename_map(self, old_ws:ROOT.RooWorkspace, new_ws:ROOT.RooWorkspace):
        rename_map = {}
        #regex = re.compile(r"\(([\w=,]+)\)")
        regex = re.compile(r"\((.+)\)")
        strict = self.config['strict']
        # rename definitions from the "Map" node
        for expr in self.actions['map']:
            match = regex.search(expr)
            if not match:
                raise RuntimeError(f"invalid expression for a map: {expr}")
            rename_str = [i for i in match.group(1).split(",") if i]
            if len(rename_str) < 2:
                raise RuntimeError(f"no members found for a map: {expr}")
            rename_str = rename_str[1:]
            for item in rename_str:
                tokens = item.split("=")
                if len(tokens) != 2:
                    raise RuntimeError(f"invalid rename syntax: {item}")
                old_name = tokens[0]
                new_name = tokens[1]
                if not old_ws.arg(old_name):
                    if strict:
                        raise RuntimeError(f"object {old_name} (-> {new_name}) does not exist in the original workspace")
                    else:
                        self.stdout.warning(f"object {old_name} does not exist in the original workspace, skipping...")
                        continue
                if not new_ws.arg(new_name):
                    raise RuntimeError(f"object {new_name} (<- {old_name}) does not exist in the temporary workspace")
                if old_name in rename_map:
                    raise RuntimeError(f"object {old_name} is renamed more than once")
                rename_map[old_name] = new_name
        # rename definitions from the "Rename" node
        for old_name, new_name in self.actions['rename']['variable'].items():
            if not old_ws.arg(old_name):
                if strict:
                    raise RuntimeError(f"object {old_name} (-> {new_name}) does not exist int the original workspace")
                else:
                    self.stdout.warning(f"object {old_name} does not exist in the original workspace, skipping...")
                    continue
            if old_name in rename_map:
                raise RuntimeError(f"object {old_name} is renamed more than once")
            rename_map[old_name] = new_name
        return rename_map
    
    def remake_simultaneous_pdf(self, old_mc:ROOT.RooStats.ModelConfig, new_ws:ROOT.RooWorkspace):
        pdf_name = old_mc.GetPdf().GetName()
        pdf = new_ws.pdf(pdf_name)
        if not pdf:
            raise RuntimeError(f"pdf {pdf_name} does not exist in workspace {new_ws.GetName()}")
        category = pdf.indexCat()
        n_cat = len(category)
        
        observables = self.extract_ws_variables(new_ws, old_mc.GetObservables(), True)
        nuisance_parameters = self.extract_ws_variables(new_ws, old_mc.GetNuisanceParameters(), False)

        pdf_map = {}
        
        for i in range(n_cat):
            category.setBin(i)
            category_name = category.getLabel()
            self.stdout.info(f"INFO: Creating new pdf for the category {category_name}")
            pdf_i = pdf.getPdf(category_name)
            
            base_components = ROOT.RooArgSet()
            obs_terms = ROOT.RooArgList()
            dis_constraints = ROOT.RooArgList()
            ROOT.RooStats.FactorizePdf(observables, pdf_i, obs_terms, dis_constraints)
            base_components.add(obs_terms)
            
            # remove constraint pdfs that are no longer needed
            if dis_constraints.getSize() > 0:
                constraints = pdf_i.getAllConstraints(observables, nuisance_parameters, True)
                base_components.add(constraints)
            
            new_pdf_name = f"{pdf_i.GetName()}__addConstr"
            # adding additional constraint pdf
            for i, item in enumerate(self.actions['constraint']):
                pdf_name = item['pdf']
                nuis_names = item['nuis']
                glob_names = item['glob']
                independent = item['independent']
                new_pdf = new_ws.pdf(pdf_name)
                if not new_pdf:
                    raise RuntimeError(f"pdf {pdf_name} does not exist in the new workspace")
                if independent:
                    base_components.add(new_pdf)
                    self.stdout.info(f"INFO: Adding independent constraint pdf \"{pdf_name}\" to "
                                     f"\"{new_pdf_name}\"")
                    continue
                # check whether the current category depends on the constraint pdf
                for nuis_name in nuis_names:
                    nuis_var = self._get_relevant_variable(nuis_name, new_ws, pdf_i)
                    if (nuis_var is not None):
                        base_components.add(new_pdf)
                        break
                        
            
            pdf_map[category_name] = ROOT.RooProdPdf(new_pdf_name, new_pdf_name, base_components)
        c_pdf_map = ExtendedModel.get_pdf_map(pdf_map)
        pdf_name = old_mc.GetPdf().GetName()
        sim_pdf = ROOT.RooSimultaneous(pdf_name, f"{pdf_name}__addConstr", c_pdf_map, category)
        return sim_pdf
    
    @staticmethod
    def _get_relevant_variable(var_name:str, ws:ROOT.RooWorkspace, pdf:ROOT.RooAbsPdf, warn:bool=False):
        var = ws.var(var_name)
        if not var:
            if warn:
                self.stdout.warning(f"WARNING: Variable {var_name} does not exist in the new workspace")
            return None
        if not pdf.dependsOn(var):
            if warn:
                self.stdout.warning(f"WARNING: Variable {var_name} exists in the new workspace but is "
                                    f"not part of the provided pdf {pdf.GetName()}")
            return None
        return var
        
    def implement_flexible_interp_var(self, ws:ROOT.RooWorkspace, expr:str):
        # parse attributes
        expr = re.sub('\s+', '', expr)
        program = re.compile(r"FlexibleInterpVar::(?P<name>[\w]+)\((?P<NPName>[\w,]+),"
                             r"(?P<nominal>[\w,]+),(?P<errHi>[\w,]+),(?P<errLo>[\w,]+),"
                             r"(?P<interpolation>[\w,]+)\)")
        result = program.search(expr)
        if not result:
            raise RuntimeError(f"invalid format for FlexibleInterpVar definition: {expr}")
        groups = result.groupdict()
        response_name = groups['name']
        nuis_name     = groups['NPName']
        nominal       = float(groups['nominal'])
        error_hi      = float(groups['errHi'])
        error_lo      = float(groups['errLo'])
        interpolation = int(groups['interpolation'])
        
        sigma_var_low = np.array([nominal + error_lo])
        sigma_var_high = np.array([nominal + error_hi])
        code = np.array([interpolation])
        
        nuis_arglist = ROOT.RooArgList()
        nuis = self.get_workspace_arg(ws, nuis_name)
        nuis_arglist.add(nuis)
        
        function = ROOT.RooStats.HistFactory.FlexibleInterpVar(response_name, response_name, nuis_arglist,
                                                               nominal, sigma_var_low, sigma_var_high, code)
        self.import_object(ws, function)
        self.stdout.info(f"INFO: Implemented FlexibleInterpVar \"{response_name}\"")
        self.stdout.info(str(py_V))

    
    def implement_multi_var_gaussian(self, ws:ROOT.RooWorkspace, expr:str):
        # parse attributes
        expr = re.sub('\s+', '', expr)
        program = re.compile(r"RooMultiVarGaussian::(?P<name>[\w]+)\({(?P<obsList>[\w,]+)}:"
                             r"{(?P<meanList>[\w,]+)}:{(?P<uncertList>[\w,]+)}:"
                             r"{(?P<correlationList>[\w,]+)}\)")
        result = program.search(expr)
        if not result:
            raise RuntimeError(f"invalid format for RooMultiVarGaussian definition: {expr}")
        groups = result.groupdict()
        function_name    = groups['name']
        obs_list         = groups['obsList'].split(",")
        mean_list        = groups['meanList'].split(",")
        uncert_list      = groups['uncertList'].split(",")
        correlation_list = groups['correlationList'].split(",")
        
        n_poi = len(obs_list)
        if not all(len(l) == n_poi for l in [obs_list, mean_list, uncert_list]):
            raise RuntimeError(f"number of items in each argument group of a RooMultiVarGaussian "
                               f"definition must be equal: {expr}")
        if len(correlation_list) != (n_poi * (n_poi - 1) // 2):
            raise RuntimeError("number of correlation matrix elements must be N * (N-1) / 2, where "
                               "N is the dimension of the matrix")
        obs_arglist = ROOT.RooArgList()
        mean_arglist = ROOT.RooArgList()
        for i in range(n_poi):
            obs_name = obs_list[i]
            mean_name = mean_list[i]
            obs = self.get_workspace_arg(ws, obs_name)
            obs_arglist.add(obs)
            mean = self.get_workspace_arg(ws, mean_name)
            mean_arglist.add(mean)
        V = ROOT.TMatrixDSym(n_poi)
        py_V = np.zeros(n_poi * n_poi).reshape(n_poi, n_poi)
        index = 0
        for i in range(n_poi):
            for j in range(i, n_poi):
                if (i == j):
                    v = float(uncert_list[i])
                    V[i, i] = v * v
                    py_V[i, i] = v * v
                elif (i < j):
                    v = float(correlation_list[index]) * float(uncert_list[i]) * float(uncert_list[j])
                    V[i, j] = v
                    V[j, i] = v
                    py_V[i, j] = v
                    py_V[j, i] = v
                    index += 1
        function = RooMultiVarGaussian(function_name, function_name, obs_arglist, mean_arglist, V)
        self.import_object(ws, function)
        self.stdout.info(f"INFO: Implemented RooMultiVarGaussian \"{function_name}\" with correlation matrix")
        self.stdout.info(str(py_V))


from typing import Optional

from .rooproc_helper_action import RooProcHelperAction

class RooProcTreeName(RooProcHelperAction):
    
    def __init__(self, treename:str):
        super().__init__(treename=treename)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        return cls(treename=main_text)
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        treename = params['treename']
        processor.treename = treename
        return processor


from typing import Optional, Union, List, Dict
import os
import re
import json
import time

import numpy as np

import ROOT
from ROOT.RooPrintable import (kName, kClassName, kValue, kArgs, kExtras, kAddress,
                               kTitle, kCollectionHeader, kSingleLine)

import quickstats
from quickstats import semistaticmethod, Timer, AbstractObject, GeneralEnum
from quickstats.components import ExtendedModel
from quickstats.utils.io import Verbosity, get_colored_text, format_comparison_text
from quickstats.utils.root_utils import load_macro, get_macro_dir
from quickstats.maths.numerics import is_float, pretty_value
from quickstats.components.basics import WSArgument
from quickstats.interface.root.roofit_extension import get_str_data

kUnique        = 0b000001
kRedefined     = 0b000010
kReconstituted = 0b000100
kRenamed       = 0b001000
kIdentical     = 0b010000
kRegrouped     = 0b100000

class WSItemType(GeneralEnum):
    
    PDF      = (0, kClassName|kName|kArgs , kSingleLine)
    FUNCTION = (1, kClassName|kName|kArgs , kSingleLine)
    VARIABLE = (2, kName|kValue|kExtras   , kSingleLine)
    OTHER    = (3, 0                      , kSingleLine)
    
    def __new__(cls, value:int, default_content:int, default_style:int):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.default_content = default_content
        obj.default_style = default_style
        return obj
    
class WSItem(GeneralEnum):
    
    WORKSPACE                        = (0,  WSItemType.OTHER, 'Workspace', None, False, False, 0b11001)
    CATEGORY                         = (1,  WSItemType.OTHER, 'Categories', None, True, False, 0b10011)
    SNAPSHOT                         = (2,  WSItemType.OTHER, 'Snapshots', None, False, False, 0b10011)
    DATASET                          = (3,  WSItemType.OTHER, 'Datasets', None, False, False, 0b11011)
    PDF                              = (4,  WSItemType.PDF, 'Pdfs', None, True, True, 0b01011)
    FUNCTION                         = (5,  WSItemType.FUNCTION, 'Functions', None, True, True, 0b01011)
    POI                              = (6,  WSItemType.VARIABLE, 'POIs', WSArgument.POI, True, False, 0b01011)
    NUISANCE_PARAMETER               = (7,  WSItemType.VARIABLE, 'Nuisance Parameters',
                                        WSArgument.NUISANCE_PARAMETER, True, False, 0b01011)
    GLOBAL_OBSERVABLE                = (8,  WSItemType.VARIABLE, 'Global Observables',
                                        WSArgument.GLOBAL_OBSERVABLE, True, False, 0b01011)    
    CONSTRAINED_NUISANCE_PARAMETER   = (9,  WSItemType.VARIABLE, 'Constrained Nuisance Parameters',
                                        WSArgument.CONSTRAINED_NUISANCE_PARAMETER, True, False, 0b01011)
    UNCONSTRAINED_NUISANCE_PARAMETER = (10, WSItemType.VARIABLE, 'Unconstrained Nuisance Parameters',
                                        WSArgument.UNCONSTRAINED_NUISANCE_PARAMETER, True, False, 0b01011)
    AUXILIARY                        = (11, WSItemType.VARIABLE, 'Auxiliary Variables',
                                        WSArgument.AUXILIARY, True, False, 0b01011)
    
    def __new__(cls, value:int, item_type:WSItemType, title:str, var_repr:Optional[WSArgument],
                support_definition:bool, is_composite:bool, default_visibility:int):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.item_type = item_type
        obj.title     = title
        obj.var_repr  = var_repr
        obj.support_definition = support_definition
        obj.is_composite = is_composite
        obj.default_visibility = default_visibility
        return obj
            
class ComparisonData:

    def __init__(self, title:str="", support_definition:bool=True, is_composite:bool=False):
        self.title = title
        self.support_definition = support_definition
        self.is_composite = is_composite
        self.reset()
        
    @classmethod
    def from_WSItem(cls, item:WSItem):
        title = item.title
        support_definition = item.support_definition
        is_composite = item.is_composite
        instance = cls(title=title, support_definition=support_definition, is_composite=is_composite)
        return instance
    
    @classmethod
    def from_dataframe(cls, item:WSItem, df:"pandas.DataFrame"):
        instance = cls.from_WSItem(item)
        instance.df_merge = df
        instance._restore_raw_df()
        return instance
    
    def reset(self):
        import pandas as pd
        self.df_1 = pd.DataFrame({"name":[], "class":[], "definition":[]})
        self.df_2 = pd.DataFrame({"name":[], "class":[], "definition":[]})
        self.df_merge = None
        self.df_factorized = None
        self.mapping_statistics_left = None
        self.mapping_statistics_right = None

    def set_title(self, title:str):
        self.title = title
        
    def add_single(self, name_1:Optional[str]=None, name_2:Optional[str]=None,
                   definition_1:str="", definition_2:str="",
                   class_1:str="", class_2:str=""):
        import pandas as pd
        if name_1 is not None:
            df_1 = pd.DataFrame.from_records([{"name": name_1, "class": "", "definition": definition_1}])
            self.df_1 = pd.concat([self.df_1, df_1], ignore_index=True)
        if name_2 is not None:
            df_2 = pd.DataFrame.from_records([{"name": name_2, "class": "", "definition": definition_2}])
            self.df_2 = pd.concat([self.df_2, df_2], ignore_index=True)
        
    def add(self, df_1:Optional["pandas.DataFrame"]=None, df_2:Optional["pandas.DataFrame"]=None):
        import pandas as pd
        if df_1 is not None:
            self.df_1 = pd.concat([self.df_1, df_1], ignore_index=True)
        if df_2 is not None:
            self.df_2 = pd.concat([self.df_2, df_2], ignore_index=True)
            
    def factorize_definition(self):
        regex = re.compile(r"([a-zA-Z_][a-zA-Z0-9_]*)")
        df = self.get_df("redefined")
        args_left = df['definition_left'].str.findall(regex)
        args_right = df['definition_right'].str.findall(regex)
        df.loc[:, ['args_left']] = args_left
        df.loc[:, ['args_right']] = args_right
        df = df[['name', 'definition_left', 'definition_right', 'args_left', 'args_right']]
        df['nargs_left'] = df['args_left'].apply(len)
        df['nargs_right'] = df['args_right'].apply(len)
        df['same_nargs'] = df['nargs_left'] == df['nargs_right']
        self.df_factorized = df
    
    def evaluate_possible_mappings(self):
        if self.df_factorized is None:
            self.factorize_definition()
        target_df = self.df_factorized[self.df_factorized["same_nargs"]]
        a = target_df['args_left'].values
        b = target_df['args_right'].values
        names = target_df['name'].values
        # left to right mapping
        statistics_left = {}
        # right to left mapping
        statistics_right = {}
        size = len(a)
        for i in range(size):
            name_i = names[i]
            size_i = len(a[i])
            for j in range(size_i):
                a_ij = a[i][j]
                b_ij = b[i][j]
                if a_ij != b_ij:
                    if a_ij not in statistics_left:
                        statistics_left[a_ij] = {}
                    if b_ij not in statistics_right:
                        statistics_right[b_ij] = {}
                    if b_ij not in statistics_left[a_ij]:
                        statistics_left[a_ij][b_ij] = 0
                    if a_ij not in statistics_right[b_ij]:
                        statistics_right[b_ij][a_ij] = 0
                    statistics_left[a_ij][b_ij] += 1
                    statistics_right[b_ij][a_ij] += 1
        self.mapping_statistics_left = statistics_left
        self.mapping_statistics_right = statistics_right
    
    @staticmethod
    def _get_combined_mapping_statistics(all_mappings:List[Dict]):
        combined_mapping = {}
        for mapping in all_mappings:
            for key, candidate_counts in mapping.items():
                if key not in combined_mapping:
                    combined_mapping[key] = {}
                for candidate, count in candidate_counts.items():
                    if candidate not in combined_mapping[key]:
                        combined_mapping[key][candidate] = 0
                    combined_mapping[key][candidate] += count
        return combined_mapping
    
    def process_mappings(self, mappings:Dict):
        if self.df_merge is None:
            self.process()
        df = self.df_merge
        df.set_index("name", inplace=True)
        for name_1, name_2 in mappings['one_to_one'].items():
            if name_1 in df.index:
                series_1 = df.loc[name_1]
                if series_1['_merge'] != "left_only":
                    continue
                # object type did not change
                if name_2 in df.index:
                    series_2 = df.loc[name_2]
                    definition_1 = series_1['definition_left']
                    # remove original name by the new name in the definition
                    definition_1 = re.sub(r"\b%s\b" % name_1, name_2, definition_1)
                    definition_2 = series_2['definition_right']
                    df.loc[name_1, ["name_alt"]] = name_2
                    df.loc[name_2, ["name_alt"]] = name_1
                    # different name, same definition
                    if definition_1 == definition_2:
                        df.loc[name_1, ["category"]] = "renamed"
                        df.loc[name_2, ["category"]] = "renamed"
                    # different name, different definition
                    else:
                        df.loc[name_1, ["category"]] = "remapped"
                        df.loc[name_2, ["category"]] = "remapped"
                    continue
                else:
                    # object mapped to object of another type
                    df.loc[name_1, ["name_alt"]] = name_2
                    df.loc[name_1, ["category"]] = "remapped"
            # object of another type mapped to object of current type
            elif name_2 in df.index:
                if df.loc[name_2]['_merge'] != "right_only":
                    continue
                # object mapped to object of another type
                df.loc[name_2, ["name_alt"]] = name_1
                df.loc[name_2, ["category"]] = "remapped"

        for name_1, names_2 in mappings['split'].items():
            if name_1 in df.index:
                if df.loc[name_1]['_merge'] != "left_only":
                    continue
                df.loc[name_1, ["name_alt"]] = "::".join(names_2)
                df.loc[name_1, ["category"]] = "splitted"
            for name_2 in names_2:
                # object type did not change
                if name_2 not in df.index:
                    continue
                if df.loc[name_2]['_merge'] != "right_only":
                    continue
                df.loc[name_2, ["name_alt"]] = name_1
                df.loc[name_2, ["category"]] = "splitted"

        for name_2, names_1 in mappings['merge'].items():
            for name_1 in names_1:
                if name_1 not in df.index:
                    continue
                if df.loc[name_1]['_merge'] != "left_only":
                    continue
                df.loc[name_1, ["name_alt"]] = name_2
                df.loc[name_1, ["category"]] = "merged"
            if name_2 in df.index:
                if df.loc[name_2]['_merge'] != "right_only":
                    continue
                df.loc[name_2, ["name_alt"]] = "::".join(names_1)
                df.loc[name_2, ["category"]] = "merged"
                    
        df.reset_index(inplace=True)
        
    def process(self):
        df_merge = self.df_1.merge(self.df_2, on="name", how='outer', indicator=True, sort=True,
                                   suffixes=('_left', '_right'), validate="one_to_one")
        df_merge['same_definition'] = df_merge['definition_left'] == df_merge['definition_right']
        columns = ["name", "definition_left", "definition_right", "class_left", "class_right"]
        import pandas as pd
        df_merge[columns] = df_merge[columns].astype(str)
        regex = re.compile(r"([a-zA-Z_][a-zA-Z0-9_]*)")
        expression_left  = df_merge['definition_left'].str.replace(regex, "")
        expression_right = df_merge['definition_right'].str.replace(regex, "")
        df_merge['same_expression'] = expression_left == expression_right
        df_merge['category'] = df_merge['_merge']
        new_categories = ["identical", "redefined", "reconstituted", "renamed", "remapped", "splitted", "merged"]
        df_merge['category'] = df_merge['category'].cat.add_categories(new_categories)
        mask_identical = (df_merge['_merge'] == "both") & (df_merge['same_definition'])
        mask_redefined = (df_merge['_merge'] == "both") & (~df_merge['same_definition'])
        df_merge.loc[mask_identical, ['category']] = 'identical'
        df_merge.loc[mask_redefined, ['category']] = 'redefined'
        df_merge['name_alt'] = None
        self.df_merge = df_merge
        if self.is_composite:
            self.factorize_definition()
            self.evaluate_possible_mappings()
        
    def get_df(self, category:str="any"):
        """
            Return a copy of the core dataframe.
            Arguments:
                category: str
                    Filter the dataframe by criteria. Allowed choices:
                    identical     = same name, same definition
                    redefined     = same name, different definition
                    reexpressed   = same name, different expression
                    reconstituted = same name, same expression, different constituents
                    left_only     = different name, found in left only
                    right_only    = different name, found in right only
                    unique_left   = different name, found in left only, no common server with right
                    unique_right  = different name, found in right only, no common server with right                   
                    renamed       = different name, same definition, same servers
                    remapped      = different name, different definition, same servers
                    splitted      = a single instance of left plays the role of multiple instances of right
                    merged        = roles of multiple instances of left played by a single instance of right
                    common        = same name
                    distinct      = different_name
                    left          = everything found in left
                    right         = everything found in left
                    any           = everything
        """
        if self.df_merge is None:
            self.process()
        if category == "any":
            df = self.df_merge
        elif category in ["common", "identical", "redefined", "reexpressed", "reconstituted"]:
            df = self.df_merge[self.df_merge['_merge'] == "both"]
            if category == "identical":
                df = df[df['category'] == "identical"]
            elif category in ["redefined", "reexpressed", "reconstituted"]:
                df = df[df['category'] == "redefined"]
                if category == "reexpressed":
                    df = df[~df['same_expression']]
                if category == "reconstituted":
                    df = df[df['same_expression']]
        elif category in ["distinct", "renamed", "remapped", "splitted", "merged"]:
            df = self.df_merge[self.df_merge['_merge'] != "both"]
            if category == "renamed":
                df = df[df['category'] == "renamed"]
            elif category == "remapped":
                df = df[df['category'] == "remapped"]
            elif category == "splitted":
                df = df[df['category'] == "splitted"]
            elif category == "merged":
                df = df[df['category'] == "merged"]                
        elif category == "left_only":
            df = self.df_merge[self.df_merge['_merge'] == "left_only"]
        elif category == "right_only":
            df = self.df_merge[self.df_merge['_merge'] == "right_only"]
        elif category == "unique_left":
            df = self.df_merge[self.df_merge['category'] == "left_only"]
        elif category == "unique_right":
            df = self.df_merge[self.df_merge['category'] == "right_only"]
        elif category == "left":
            df = self.df_merge[self.df_merge['_merge'] != "right_only"]
        elif category == "right":
            df = self.df_merge[self.df_merge['_merge'] != "left_only"]
        else:
            raise ValueError('invalid category (choose from "identical", "redefined", "reconstituted", '
                             '"reexpressed", "left_only", "right_only", "unique_left", "unique_right", '
                             '"renamed", "remapped", "splitted", "merged", "common", "distinct", "any")')
        # make a copy and reset index
        df = df.reset_index(drop=True)
        return df
    
    def _restore_raw_df(self):
        if self.df_merge is None:
            raise RuntimeError("no merged dataframe to restore")
        df = self.df_merge
        df_1 = df[df['_merge'] != "right_only"][['name','class_left', 'definition_left']]
        df_1 = df_1.rename(columns={"class_left":"class", "definition_left":"defintion"})
        df_2 = df[df['_merge'] != "left_only"][['name','class_right', 'definition_right']]
        df_2 = df_1.rename(columns={"class_right":"class", "definition_right":"defintion"})
        self.df_1 = df_1
        self.df_2 = df_2
        if self.is_composite:
            self.factorize_definition()
            self.evaluate_possible_mappings()
            
    def _get_content_summary_str(self, contents:List[str], title:str, title_color:str,
                                 content_color:str, indent:str="   ", show_content:bool=True):
        summary_str = ""
        size = len(contents)
        if size > 0:
            s = f"{indent}[{title} ({size})]\n"
            summary_str += get_colored_text(s, title_color)
            if not show_content:
                return summary_str
            for content in contents:
                s = f"{indent*2}{content}\n"
                summary_str += get_colored_text(s, content_color)
        return summary_str
    
    def _get_mapped_content_summary_str(self, contents_1:List[str], contents_2:List[str],
                                        title:str, title_color:str,
                                        equal_color:Optional[str]=None,
                                        delete_color:str="red",
                                        insert_color:str="green",
                                        indent:str="   ",
                                        show_content:bool=True):
        summary_str = ""
        size_1 = len(contents_1)
        size_2 = len(contents_2)
        if size_1 != size_2:
            raise ValueError("content_1 and content_2 must have the same size")
        if size_1 > 0:
            s = f"{indent}[{title} ({size_1})]\n"
            summary_str += get_colored_text(s, title_color)
            if not show_content:
                return summary_str
            for (content_1, content_2) in zip(contents_1, contents_2):
                s_left, s_right = format_comparison_text(content_1, content_2,
                                                         equal_color, delete_color, insert_color)
                summary_str += f"{indent*2}{s_left} -> {s_right}\n"
        return summary_str
    
    def _get_redef_mapping_contents(self, df:"pandas.DataFrame"):
        contents = df[['definition_left', 'definition_right']].to_dict('list')
        contents_1 = contents["definition_left"]
        contents_2 = contents["definition_right"]
        return contents_1, contents_2
    
    def _get_ext_mapping_contents(self, df:"pandas.DataFrame", ext_ref:Optional[Dict]=None):
        if ext_ref is None:
            ext_ref = {}
        df_1 = df[df["_merge"] == "left_only"]
        df_2 = df[df["_merge"] == "right_only"]
        orig_names_1 = df_1['name'].values
        orig_names_2 = df_2['name'].values
        alt_names_1 = df_1["name_alt"].values
        alt_names_2 = df_2["name_alt"].values
        mapping_1 = dict(zip(orig_names_1, alt_names_1))
        mapping_2 = dict(zip(orig_names_2, alt_names_2))
        df_1.set_index('name', inplace=True)
        df_2.set_index('name', inplace=True)
        contents_1 = []
        contents_2 = []
        # cases:
        #[{"LA": "RA1"}, {"LA": "RB"}, {"LA" : "RB1::RB2::RB3"}, {"LA": "RA1::RA2::RA3"}, {"LA": "RB1::RA1"}
        # {"RA": "LA1"}, {"RA": "LB"}, {"RA": "LA1::LA2::LA3"}, {"RA": "LB1::LB2::LB3"}, {"RA": "LB1::LA1"}]
        for name, alt_names in mapping_1.items():
            contents_1.append(df_1.loc[name]["definition_left"])
            content_components = []
            alt_names = alt_names.split("::")
            for alt_name in alt_names:
                # left same item type -> right same item type;
                if (alt_name in mapping_2):
                    content_components.append(df_2.loc[alt_name]["definition_right"])
                # left same item type -> right diff item type;
                elif (alt_name  in ext_ref):
                    content_components.append(ext_ref[alt_name])
                else:
                    content_components.append(f"{alt_name} [definition not inferred]")
            contents_2.append(" & ".join(content_components))
        for name, alt_names in mapping_2.items():
            # left same item type -> right same item type; already dealt with
            if ("::" not in alt_names) and (alt_names in mapping_1):
                continue
            contents_2.append(df_2.loc[name]["definition_right"])
            content_components = []
            alt_names = alt_names.split("::")
            for alt_name in alt_names:
                # left diff item type -> right same item type; 
                if (alt_name  in ext_ref):
                    content_components.append(ext_ref[alt_name])
                else:
                    content_components.append(f"{alt_name} [definition not inferred]")
            contents_1.append(" & ".join(content_components))
        return contents_1, contents_2
    
    def get_summary_str(self, visibility:int=0b00011, indent:str="   ",
                        ext_ref:Optional[Dict]=None):
        """
            Arguments:
                visibility: integer
                    boolean mask for showing definitions of certain objects
                    0b000001 = show definitions for unique objects
                    0b000010 = show definitions for redefined objects
                    0b000100 = show definitions for reconstituted objects
                    0b001000 = show definitions for renamed/remapped objects
                    0b010000 = show definitions for identical objects
                    0b010000 = show definitions for regroupped objects
        """
        if ext_ref is None:
            ext_ref = {}
        s = f"{self.title}:\n"
        summary_str = get_colored_text(s, "bright magenta")
        if self.support_definition:
            df = self.get_df("identical")
            content_kwargs = {
                "contents": df['definition_left'].values,
                "title": "Common Object (Identical Definition)",
                "title_color": "bright yellow",
                "content_color": "green",
                "indent": indent,
                "show_content": visibility & kIdentical
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
            if self.is_composite:
                df = self.get_df("reconstituted")
                contents_1, contents_2 = self._get_redef_mapping_contents(df)
                content_kwargs = {
                    "contents_1": contents_1,
                    "contents_2": contents_2,
                    "title": "Common Object (Same Expression, Modified Members)",
                    "title_color": "bright yellow",
                    "indent": indent,
                    "show_content": visibility & kReconstituted
                }
                summary_str += self._get_mapped_content_summary_str(**content_kwargs)
                df = self.get_df("reexpressed")
                contents_1, contents_2 = self._get_redef_mapping_contents(df)
                content_kwargs = {
                    "contents_1": contents_1,
                    "contents_2": contents_2,
                    "title": "Common Object (Modified Expression)",
                    "title_color": "bright yellow",
                    "indent": indent,
                    "show_content": visibility & kRedefined
                }
                summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            else:
                df = self.get_df("redefined")
                contents_1, contents_2 = self._get_redef_mapping_contents(df)
                content_kwargs = {
                    "contents_1": contents_1,
                    "contents_2": contents_2,
                    "title": "Common Object (Modified Definition)",
                    "title_color": "bright yellow",
                    "indent": indent,
                    "show_content": visibility & kRedefined
                }
                summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            
            df = self.get_df("renamed")
            contents_1, contents_2 = self._get_ext_mapping_contents(df, ext_ref)
            content_kwargs = {
                "contents_1": contents_1,
                "contents_2": contents_2,
                "title": "Renamed Object (Modified Name, Same Definition)",
                "title_color": "bright yellow",
                "indent": indent,
                "show_content": visibility & kRenamed
            }
            summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            
            df = self.get_df("remapped")
            contents_1, contents_2 = self._get_ext_mapping_contents(df, ext_ref)
            content_kwargs = {
                "contents_1": contents_1,
                "contents_2": contents_2,
                "title": "Remapped Object (Modified Name, Modified Definition, One to One)",
                "title_color": "bright yellow",
                "indent": indent,
                "show_content": visibility & kRenamed
            }
            summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            
            df = self.get_df("splitted")
            contents_1, contents_2 = self._get_ext_mapping_contents(df, ext_ref)
            content_kwargs = {
                "contents_1": contents_1,
                "contents_2": contents_2,
                "title": "Splitted Object (Modified Name, Modified Definition, One to Many)",
                "title_color": "bright yellow",
                "indent": indent,
                "show_content": visibility & kRegrouped
            }
            summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            
            df = self.get_df("merged")
            contents_1, contents_2 = self._get_ext_mapping_contents(df, ext_ref)
            content_kwargs = {
                "contents_1": contents_1,
                "contents_2": contents_2,
                "title": "Merged Object (Modified Name, Modified Definition, Many to One)",
                "title_color": "bright yellow",
                "indent": indent,
                "show_content": visibility & kRegrouped
            }
            summary_str += self._get_mapped_content_summary_str(**content_kwargs)            
            
            df = self.get_df("unique_left")
            content_kwargs = {
                "contents": df["definition_left"].values,
                "title": "Unique Object (Left)",
                "title_color": "bright yellow",
                "content_color": "red",
                "indent": indent
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
            
            df = self.get_df("unique_right")
            content_kwargs = {
                "contents": df["definition_right"].values,
                "title": "Unique Object (Right)",
                "title_color": "bright yellow",
                "content_color": "red",
                "indent": indent
            }            
            summary_str += self._get_content_summary_str(**content_kwargs)
        else:
            content_kwargs = {
                "contents": self.get_df("identical")['name'].values,
                "title": "Common Object",
                "title_color": "bright yellow",
                "content_color": "green",
                "indent": indent,
                "show_content": visibility & kIdentical
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
            content_kwargs = {
                "contents": self.get_df("redefined")['name'].values,
                "title": "Common Object (Modified Definition)",
                "title_color": "bright yellow",
                "content_color": "green",
                "indent": indent,
                "show_content": visibility & kRedefined
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
            df = self.get_df("renamed")
            df = df[df["_merge"] == "left_only"]
            content_kwargs = {
                "contents_1": df["name"].values,
                "contents_2": df["name_alt"].values,
                "title": "Renamed Object (Modified Name, Same Definition)",
                "title_color": "bright yellow",
                "indent": indent,
                "show_content": visibility & kRenamed
            }
            summary_str += self._get_mapped_content_summary_str(**content_kwargs)
            content_kwargs = {
                "contents": self.get_df("unique_left")['name'].values,
                "title": "Unique Object (Left)",
                "title_color": "bright yellow",
                "content_color": "red",
                "indent": indent,
                "show_content": visibility & kUnique
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
            content_kwargs = {
                "contents": self.get_df("unique_right")['name'].values,
                "title": "Unique Object (Right)",
                "title_color": "bright yellow",
                "content_color": "red",
                "indent": indent,
                "show_content": visibility & kUnique
            }
            summary_str += self._get_content_summary_str(**content_kwargs)
        return summary_str

class WSComparer(AbstractObject):
    
    kDefaultItems = [WSItem.WORKSPACE, WSItem.CATEGORY, WSItem.DATASET, WSItem.POI,
                     WSItem.PDF, WSItem.FUNCTION, WSItem.NUISANCE_PARAMETER,
                     WSItem.GLOBAL_OBSERVABLE, WSItem.AUXILIARY]
    
    @property
    def target_items(self) -> Union[List[str], List[WSItem]]:
        return self._target_items
    
    @target_items.setter
    def target_items(self, other:Optional[Union[List[str], List[WSItem], WSItem, str]]=None) -> None:
        self._target_items = self._parse_items(other)
        
    @property
    def visibility_map(self) -> Dict[str, int]:
        return self._visibility_map
    
    @visibility_map.setter
    def visibility_map(self, source:Optional[Union[str, Dict]]=None):
        if source is None:
            self._visibility_map = {}
        elif isinstance(source, dict):
            visibility_map = {}
            for key, value in source.items():
                item = WSItem.parse(key)
                item_str = item.name.lower()
                if isinstance(value, str):
                    value = int(value, 2)
                visibility_map[item_str] = value
            self._visibility_map = visibility_map
        elif isinstance(source, str):
            expressions = source.split(",")
            visibility_map = {}
            for expression in expressions:
                tokens = expression.split("=")
                if len(tokens) != 2:
                    raise ValueError("invalid string format for setting visibility")
                item_str = tokens[0]
                value = tokens[1]
                visibility_map[item_str] = value
            self.visibility_map = visibility_map
        else:
            raise ValueError("unknown argument type for setting visibility")
                       
    def __init__(self, ws_path_1:Optional[str]=None, ws_path_2:Optional[str]=None,
                 items:Optional[Union[List[str], List[WSItem], WSItem, str]]=None,
                 visibility_map:Optional[Union[str, Dict]]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        
        super().__init__(verbosity=verbosity)
        
        if (ws_path_1 is not None) and (ws_path_2 is not None):
            self.set_targets(ws_path_1, ws_path_2)
        else:
            self.model_1 = None
            self.model_2 = None
            
        self.item_contents = {
            WSItemType.PDF      : WSItemType.PDF.default_content,
            WSItemType.FUNCTION : WSItemType.FUNCTION.default_content,
            WSItemType.VARIABLE : WSItemType.VARIABLE.default_content
        }
        self.item_styles = {
            WSItemType.PDF      : WSItemType.PDF.default_style,
            WSItemType.FUNCTION : WSItemType.FUNCTION.default_style,
            WSItemType.VARIABLE : WSItemType.VARIABLE.default_style,
        }
        
        self.target_items = items
        self.visibility_map = visibility_map
        self.data = {}

    def set_targets(self, ws_path_1:str, ws_path_2:str):
        self.stdout.info(f"INFO: Initializing targets...")
        with Timer() as t:
            self.model_1 = ExtendedModel(ws_path_1, data_name=None, verbosity="WARNING")
        self.stdout.info(f'INFO: Loaded workspace (left) from "{ws_path_1}". Time taken: {t.interval:.3f}s.')
        with Timer() as t:
            self.model_2 = ExtendedModel(ws_path_2, data_name=None, verbosity="WARNING")
        self.stdout.info(f'INFO: Loaded workspace (right) from "{ws_path_2}". Time taken: {t.interval:.3f}s.')
        
    def _parse_items(self, items:Optional[Union[List[str], List[WSItem], WSItem, str]]=None) -> List[WSItem]:
        if items is None:
            parsed_items = list(self.kDefaultItems)
        elif isinstance(items, str):
            item = WSItem.parse(items)
            parsed_items = [item]
        elif isinstance(items, WSItem):
            parsed_items = [items]
        elif isinstance(items, list):
            parsed_items = [WSItem.parse(i) for i in items]
        else:
            raise ValueError(f"invalid item format for workspace comparison: {items}")
        return parsed_items
    
    def set_item_content(self, item_type:Union[str, WSItemType], content:int):
        item_type = WSItemType.parse(item_type)
        self.item_contents[item_type] = content
        
    def set_item_style(self, item_type:Union[str, WSItemType], style:int):
        item_type = WSItemType.parse(item_type)
        self.item_styles[item_type] = style
        
    def clear_data(self):
        self.data = {}
        
    def load_data(self, items:Optional[Union[List[str], List[WSItem], WSItem, str]]=None):
        
        if items is None:
            items = self.target_items
        else:
            items = self._parse_items(items)
            self.target_items = items
            
        if any(item.item_type == WSItemType.VARIABLE for item in items):
            for _item in [WSItem.PDF, WSItem.FUNCTION]:
                if _item not in items:
                    self.stdout.info(f'INFO: Added "{_item.title}" to the requested data pool needed for '
                                     'inferring renamed and redefined variables.')
                    items.append(_item)
        
        self.clear_data()
        self.stdout.info("INFO: Loading workspace information...")
        with Timer() as t:
            for item in items:
                item_name = item.name.lower()
                self.data[item_name] = self.get_item_data(item)
        self.stdout.info(f"INFO: All requested data have been successfully loaded. Time taken: {t.interval:.3f}s.")        

    def _get_workspace_data(self):
        item = WSItem.WORKSPACE
        data = ComparisonData.from_WSItem(item)
        ws_name_1 = self.model_1.workspace.GetName()
        ws_name_2 = self.model_2.workspace.GetName()
        data.add_single(name_1=ws_name_1, name_2=ws_name_2)
        return data
    
    def _get_category_data(self):
        item = WSItem.CATEGORY
        data = ComparisonData.from_WSItem(item)
        category_map_1 = self.model_1.get_category_map()
        category_map_2 = self.model_2.get_category_map()
        for category in category_map_1:
            category_map = category_map_1[category]
            definition = self.model_1._format_category_summary(category, category_map)
            data.add_single(name_1=category, definition_1=definition)
        for category in category_map_2:
            category_map = category_map_2[category]
            definition = self.model_2._format_category_summary(category, category_map)
            data.add_single(name_2=category, definition_2=definition)
        return data
    
    def _get_snapshot_data(self):
        item = WSItem.SNAPSHOT
        data = ComparisonData.from_WSItem(item)
        if (quickstats.root_version >= (6, 26, 0)):
            snapshots_1 = self.model_1.workspace.getSnapshots()
            snapshots_2 = self.model_2.workspace.getSnapshots()
            for snapshot in snapshots_1:
                data.add_single(name_1=snapshot.GetName())
            for snapshot in snapshots_2:
                data.add_single(name_2=snapshot.GetName())
        else:
            self.stdout.warning("WARNING: Snapshot listing is only available after ROOT 6.26/00")
        return data
     
    def _get_dataset_data(self):
        item = WSItem.DATASET
        data = ComparisonData.from_WSItem(item)
        datasets_1 = self.model_1.workspace.allData()
        datasets_2 = self.model_2.workspace.allData()
        for dataset in datasets_1:
            data.add_single(name_1=dataset.GetName())
        for dataset in datasets_2:
            data.add_single(name_2=dataset.GetName())
        return data
    
    def _get_pdf_data(self):
        item = WSItem.PDF
        components_1 = self.model_1.workspace.allPdfs()
        components_2 = self.model_2.workspace.allPdfs()
        data = self._get_argset_item_data(item, components_1, components_2)
        return data
    
    def _get_function_data(self):
        item = WSItem.FUNCTION
        components_1 = self.model_1.workspace.allFunctions()
        components_2 = self.model_2.workspace.allFunctions()
        data = self._get_argset_item_data(item, components_1, components_2)
        return data
    
    def _get_variable_data(self, item:WSItem):
        if item.item_type != WSItemType.VARIABLE:
            raise ValueError("item type must be variable")
        variable = item.var_repr
        components_1 = self.model_1.get_variables(variable)
        components_2 = self.model_2.get_variables(variable)
        data = self._get_argset_item_data(item, components_1, components_2)
        return data
    
    def _get_argset_item_data(self, item:WSItem, components_1:ROOT.RooArgSet, components_2:ROOT.RooArgSet):
        data = ComparisonData.from_WSItem(item)
        content = self.item_contents[item.item_type]
        style   = self.item_styles[item.item_type]
        data_1 = get_str_data(components_1, fill_classes=True,
                              fill_definitions=True,
                              content=content, style=style, fmt="dataframe")
        data_2 = get_str_data(components_2, fill_classes=True,
                              fill_definitions=True,
                              content=content, style=style, fmt="dataframe")
        data.add(data_1, data_2)
        return data
    
    def get_item_data(self, item:Union[str, WSItem]):
        item = WSItem.parse(item)
        title = item.title
        self.stdout.info(f'INFO: Retrieving information for the item: "{title}"')
        if item == WSItem.WORKSPACE:
            data = self._get_workspace_data()
        elif item == WSItem.CATEGORY:
            data = self._get_category_data()
        elif item == WSItem.SNAPSHOT:
            data = self._get_snapshot_data()
        elif item == WSItem.DATASET:
            data = self._get_dataset_data()
        elif item == WSItem.PDF:
            data = self._get_pdf_data()
        elif item == WSItem.FUNCTION:
            data = self._get_function_data()
        elif item.item_type == WSItemType.VARIABLE:
            data = self._get_variable_data(item)
        else:
            raise RuntimeError(f"unsupported item type: {item}")
        data.process()
        return data
  
    @staticmethod
    def _dataset_distribution_is_equal(dist_1:Dict, dist_2:Dict):
        categories_1 = dist_1.keys()
        categories_2 = dist_2.keys()
        if set(categories_1) != set(categories_2):
            return False
        for category in categories_1:
            x_1 = dist_1[category]['x']
            x_2 = dist_2[category]['x']
            y_1 = dist_1[category]['y']
            y_2 = dist_2[category]['y']
            if (not np.allclose(x_1, x_2)) or (not np.allclose(y_1, y_2)):
                return False
        return True
    
    def _get_dataset_dist(self, name_1:Optional[str]=None, name_2:Optional[str]=None):
        if name_1 is not None:
            dataset_1 = self.model_1.workspace.data(name_1)
            dist_1 = self.model_1.get_dataset_distributions(dataset_1)
        else:
            dist_1 = None
        if name_2 is not None:
            dataset_2 = self.model_2.workspace.data(name_2)
            dist_2 = self.model_2.get_dataset_distributions(dataset_2)
        else:
            dist_2 = None
        return dist_1, dist_2

    def _process_workspace_data(self):
        item = WSItem.WORKSPACE
        item_name = item.name.lower()
        if item_name not in self.data:
            return None
        data = self.data[item_name]
        df_common = data.get_df("common")
        if len(df_common) == 0:
            df = data.df_merge
            index_1 = df["_merge"] == "left_only"
            index_2 = df["_merge"] == "right_only"
            names_1 = df[index_1]['name'].values
            names_2 = df[index_2]['name'].values
            if (len(names_1) != 1) or (len(names_2) != 1):
                raise RuntimeError("found multiple left/right workspaces in the comparison data")
            df.loc[index_1, ['category']] = "renamed"
            df.loc[index_2, ['category']] = "renamed"
            df.loc[index_1, ['name_alt']] = names_2[0]
            df.loc[index_2, ['name_alt']] = names_1[0]
    
    def _process_dataset_data(self):
        item = WSItem.DATASET
        item_name = item.name.lower()
        self.stdout.info(f'INFO: Processing dataset definitions...')
        if item_name not in self.data:
            self.stdout.warning("WARNING: Dataset data not loaded. Skipping.")
            return None
        with Timer() as t:
            data = self.data[item_name]
            common_dataset_names = data.get_df("common")['name'].values
            dataset_names_1 = data.get_df("left_only")['name'].values
            dataset_names_2 = data.get_df("right_only")['name'].values
            df = data.df_merge
            df.set_index("name", inplace=True)
            for ds_name in common_dataset_names:
                dist_1, dist_2 = self._get_dataset_dist(ds_name, ds_name)
                same_definition = self._dataset_distribution_is_equal(dist_1, dist_2)
                if not same_definition:
                    df.loc[ds_name, ['category']] = "redefined"
            for ds_name_1 in dataset_names_1:
                candidates = []
                dist_1, _ = self._get_dataset_dist(ds_name_1, None)
                for ds_name_2 in dataset_names_2:
                    _, dist_2 = self._get_dataset_dist(None, ds_name_2)
                    same_definition = self._dataset_distribution_is_equal(dist_1, dist_2)
                    if same_definition:
                        candidates.append(ds_name_2)
                # renamed dataset
                if len(candidates) == 1:
                    ds_name_2 = candidates[0]
                    df.loc[ds_name_1, ['category']] = "renamed"
                    df.loc[ds_name_1, ['name_alt']] = ds_name_2
                    df.loc[ds_name_2, ['category']] = "renamed"
                    df.loc[ds_name_2, ['name_alt']] = ds_name_1
            df.reset_index(inplace=True)
        self.stdout.info(f"INFO: Processing finished. Time taken: {t.interval:.3f}s.")
        
    def _get_combined_mapping_statistics(self):
        pdf_item_name = WSItem.PDF.name.lower()
        function_item_name = WSItem.FUNCTION.name.lower()
        items_to_request = []
        if pdf_item_name not in self.data:
            items_to_request.append(WSItem.PDF)
        if function_item_name not in self.data:
            items_to_request.append(WSItem.FUNCTION)
        
        if len(items_to_request) > 0:
            with Timer() as t:
                for item in items_to_request:
                    self.stdout.info(f'INFO: Added "{item.title}" to the requested data pool needed for '
                                     'inferring renamed and redefined variables.')
                    item_name = item.name.lower()
                    self.data[item_name] = self.get_item_data(item)
            self.stdout.info(f"INFO: All requested data have been successfully loaded. Time taken: {t.interval:.3f}s.")
            return self._get_combined_mapping_statistics()
        
        pdf_mapping_statistics_left = self.data[pdf_item_name].mapping_statistics_left
        function_mapping_statistics_left = self.data[function_item_name].mapping_statistics_left
        pdf_mapping_statistics_right = self.data[pdf_item_name].mapping_statistics_right
        function_mapping_statistics_right = self.data[function_item_name].mapping_statistics_right        
        combined_mapping_left = ComparisonData._get_combined_mapping_statistics([pdf_mapping_statistics_left,
                                                                                 function_mapping_statistics_left])
        combined_mapping_right = ComparisonData._get_combined_mapping_statistics([pdf_mapping_statistics_right,
                                                                                  function_mapping_statistics_right])        
        return combined_mapping_left, combined_mapping_right
    
    def _get_mappings(self, mapping_statistics_left:Dict, mapping_statistics_right:Dict):
        mapping_left = {k:list(v) for k,v in mapping_statistics_left.items()}
        mapping_right = {k:list(v) for k,v in mapping_statistics_right.items()}
        mappings = {
            "one_to_one": {},
            "split": {},
            "merge": {}
        }
        for key, candidates in mapping_left.items():
            if len(candidates) == 1:
                candidate = candidates[0]
                inverse_candidates = mapping_right.get(candidate, [])
                if len(inverse_candidates) == 1 and inverse_candidates[0] == key:
                    mappings['one_to_one'][key] = candidate
                elif len(inverse_candidates) > 1:
                    # surjection for left to right mapping
                    if not all(len(mapping_left.get(icandidate, [])) == 1 for icandidate in inverse_candidates):
                        continue
                    mappings['merge'][candidate] = inverse_candidates
            else:
                # surjection for right to left mapping
                if not all(mapping_right.get(candidate, [None])[0] == key for candidate in candidates):
                    continue
                mappings['split'][key] = candidates
        return mappings
    
    def _get_items_with_definition(self):
        relevant_items = []
        for item_str in self.data:
            item = WSItem.parse(item_str)
            if item.item_type in [WSItemType.PDF, WSItemType.FUNCTION, WSItemType.VARIABLE]:
                relevant_items.append(item_str)
        return relevant_items

    def _process_renamed_and_remapped_variables(self):
        
        relevant_items = self._get_items_with_definition()
                
        if len(relevant_items) > 0:
            self.stdout.info("INFO: Processing renamed and remapped variables...")
            with Timer() as t:
                map_stat_left, map_stat_right = self._get_combined_mapping_statistics()
                mappings = self._get_mappings(map_stat_left, map_stat_right)
                for item_str in relevant_items:
                    data = self.data[item_str]
                    data.process_mappings(mappings)
            self.stdout.info(f"INFO: Processing finished. Time taken: {t.interval:.3f}s")
 
    def process_data(self):
        self._process_workspace_data()
        dataset_item_str = WSItem.DATASET.name.lower()
        if dataset_item_str in self.data:
            self._process_dataset_data()
        self._process_renamed_and_remapped_variables()
    
    def _get_definition_map(self):
        relevant_items = self._get_items_with_definition()
        definition_map = {}
        for item_str in relevant_items:
            data = self.data[item_str]
            df = data.get_df("right")
            names = df['name'].values
            definitions = df['definition_right'].values
            def_map = dict(zip(names, definitions))
            definition_map.update(def_map)
        return definition_map
    
    def get_summary_str(self, indent:str="   "):
        definition_map = self._get_definition_map()
        combined_summary_str = ""
        for item_str, data in self.data.items():
            if data.df_merge is None:
                continue
            item = WSItem.parse(item_str)
            if item not in self.target_items:
                continue
            visibility = self.visibility_map.get(item_str, item.default_visibility)
            summary_str = data.get_summary_str(visibility=visibility,
                                               ext_ref=definition_map,
                                               indent=indent)
            combined_summary_str += summary_str
        return combined_summary_str
    
    def print_summary(self, indent:str="   "):
        summary_str = self.get_summary_str(indent=indent)
        self.stdout.info(summary_str)
        
    @classmethod
    def from_json(cls, filename:str):
        with open(filename, "r") as infile:
            json_data = json.load(infile)
        instance = cls()
        items = []
        import pandas as pd
        for item_str, data in json_data.items():
            item = WSItem.parse(item_str)
            items.append(item)
            df = pd.DataFrame(data)
            instance.data[item_str] = ComparisonData.from_dataframe(item, df)
        instance.target_items = items
        return instance
    
    @classmethod
    def from_excel(cls, filename:str):
        import pandas as pd
        instance = cls()
        excel_data = pd.read_excel(filename, sheet_name=None)
        items = []
        for item_str, df in excel_data.items():
            item = WSItem.parse(item_str)
            items.append(item)
            instance.data[item_str] = ComparisonData.from_dataframe(item, df)
        instance.target_items = items
        return instance
    
    def save_json(self, filename:str):
        json_data = {}
        self.stdout.info(f'INFO: Saving comparison result in json format as "{filename}"...')
        for item_str, data in self.data.items():
            if data.df_merge is None:
                data.process()
            # only save items that are requested
            item = WSItem.parse(item_str)
            if item in self.target_items:
                json_data[item_str] = data.df_merge.to_dict('records')
        with open(filename, "w") as outfile:
            json.dump(json_data, outfile, indent=2)
    
    def save_excel(self, filename:str):
        self.stdout.info(f'INFO: Saving comparison result in excel format as "{filename}"...')
        import pandas as pd
        writer = pd.ExcelWriter(filename, engine='xlsxwriter')
        for item_str, data in self.data.items():
            if data.df_merge is None:
                data.process()
            # only save items that are requested
            item = WSItem.parse(item_str)
            df = data.df_merge
            df.to_excel(writer, sheet_name=item_str, index=False)
        writer.save()


# keywords
SELF_SYST_DOMAIN_KEYWORD = ":self:"
COMMON_SYST_DOMAIN_KEYWORD = ":common:"
NORM_SYST_KEYWORD = "yield"
SHAPE_SYST_KEYWORD = "shape"
PROCESS_KEYWORD = ":process:"
RESPONSE_KEYWORD = "response::"
OBSERVABLE_KEYWORD = ":observable:"
CATEGORY_KEYWORD = ":category:"

# logic operations
LT_OP = ":lt:"
LE_OP = ":le:"
GT_OP = ":gt:"
GE_OP = ":ge:"
AND_OP = ":and:"
OR_OP = ":or:"

# data sources
DATA_SOURCE_COUNTING = "counting"
DATA_SOURCE_ASCII = "ascii"
DATA_SOURCE_HISTOGRAM = "histogram"

# pdfs
PDF_USERDEF = "userdef"
PDF_EXTERNAL = "external"

# constraints
CONSTR_GAUSSIAN = "gaus"
CONSTR_LOGN = "logn"
CONSTR_ASYM = "asym"
CONSTR_DFD = "dfd"

# naming
RESPONSE_PREFIX = "expected__"
CONSTRTERM_PREFIX = "constr__"
GLOBALOBS_PREFIX = "RNDM__"
VARIATIONHI_PREFIX = "varHi__"
VARIATIONLO_PREFIX = "varLo__"
YIELD_PREFIX = "yield__"
PDF_PREFIX = "pdf__"
EXPECTATION_PREFIX = "expectation__"
UNCERT_HI_PREFIX = "uncertHi__"
UNCERT_LO_PREFIX = "uncertLo__"
UNCERT_SYM_PREFIX = "uncertSymm__"
SUM_PDF_NAME = "_modelSB"
FINAL_PDF_NAME = "_model"
OBS_DATASET_NAME = "obsdata"
LUMI_NAME = "_luminosity"
NORM_PREFIX = "_norm"
XS_PREFIX = "_xs"
BR_PREFIX = "_br"
EFFICIENCY_PREFIX = "_eff"
ACCEPTANCE_PREFIX = "_A"
CORRECTION_PREFIX = "_C"
COMMON_SYST_DOMAIN_NAME = "_allproc_"

#model sources
USERDEF_MODEL = "userdef"
EXTERNAL_MODEL = "external"
HISTOGRAM_MODEL = "histogram"

#blinding
RANGE_NAME_SB_LO = "SBLo"
RANGE_NAME_SB_HI = "SBHi"
RANGE_NAME_BLIND = "Blind"

# variable sets
OBS_SET = "Observables"
POI_SET = "POI"
NUIS_SET = "nuisanceParameters"
GLOB_SET = "globalObservables"
CONSTR_SET = "constraintPdfs"

# asimov generation
GEN_ASIMOV_ACTION = "genasimov"
FLOAT_ASIMOV_ACTION = "float"
RAW_ASIMOV_ACTION = "raw"
FIT_ASIMOV_ACTION = "fit"
RESET_ASIMOV_ACTION = "reset"
FIX_SYST_ASIMOV_ACTION = "fixsyst"
FIX_ALL_ASIMOV_ACTION = "fixall"
MATCH_GLOB_ASIMOV_ACTION = "matchglob"
SAVE_SNAPSHOT_ASIMOV_ACTION = "savesnapshot"

OTHER = "other"


##################################################################################################
# Author: Alkaid Cheng
# Email: chi.lung.cheng@cern.ch
##################################################################################################
import os
import sys
from typing import List, Optional, Union, Dict, Set

import numpy as np

import ROOT

import quickstats
from quickstats import semistaticmethod, AbstractObject
from quickstats.interface.root import TH1

class ExtendedRFile(AbstractObject):
    
    def __init__(self, source:Union[str, Dict[str, np.ndarray], "pandas.DataFrame"],
                 tree_name:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.initialize(source, tree_name)
        
    @property
    def file(self):
        return self._file
    
    @property
    def components(self):
        return self._components
    
    @property
    def rdf(self):
        return self._rdf
    
    @property
    def rdf_map(self):
        return self._rdf_map
    
    @property
    def cache(self):
        return self._cache
    
    @property
    def canvas(self):
        return self._canvas
    
    @staticmethod
    def is_corrupt(f:ROOT.TFile):
        if f.IsZombie():
            return True
        if f.TestBit(ROOT.TFile.kRecovered):
            return True
        if f.GetNkeys() == 0:
            return True
        return False
        
    def initialize(self, source:Union[str, Dict[str, np.ndarray]], tree_name:Optional[str]=None):
        
        self._rdf_map = {}
        self._rdf     = None
        self._canvas = None
        self._cache  = []
        self._components = []        
        
        # initialize from ROOT file
        if isinstance(source, str):
            if not os.path.exists(source):
                raise FileNotFoundError(f"workspace file {source} does not exist")
            self.stdout.info('INFO: Opening file "{}"'.format(self.fname))
            file = ROOT.TFile(source) 
            if self.is_corrupt(file):
                raise RuntimeError(f"corrupted file: {source}")
            self._file = file
            self.fname = source
            if tree_name is not None:
                self.set_tree(tree_name)
        elif isinstance(source, dict):
            self._rdf = ROOT.RDF.MakeNumpyDataFrame(source)
            self._rdf_map[None] = self._rdf
            self._file = None
            self.fname = None
        else:
            import pandas as pd
            if isinstance(source, pd.DataFrame):
                from quickstats.interface.root.data_conversion import dataframe2numpy
                source = dataframe2numpy(source)
                self.initialize(source)
            else:
                raise RuntimeError("invalid source")
        
    def create_canvas(self):
        if not self.canvas:
            self._canvas = ROOT.TCanvas()
    
    def canvas_draw(self):
        self.create_canvas()
        self.canvas.Draw()
    
    def add_tree(self, tree_name:str):
        if tree_name in self._rdf_map:
            self.stdout.info("INFO: Tree already added to the rdf collection.")
            return None
        rdf = ROOT.RDataFrame(tree_name, self.fname)
        if not rdf:
            raise RuntimeError(f"failed to load tree \"{tree_name}\"")
        self._rdf_map[tree_name] = rdf
        self.stdout.info(f"INFO: Added tree \"{tree_name}\" to the rdf collection")
        
    def set_tree(self, tree_name:str):
        if tree_name not in self._rdf_map:
            self.add_tree(tree_name)
        self._rdf = self._rdf_map[tree_name]
        self.stdout.info(f"INFO: Set active tree to \"{tree_name}\"")
    
    def validate(self):
        if self._rdf is None:
            raise RuntimeError("active tree not set")
            
    def get_Histo1D(self, name:str, n_bins:int, xmin:float, xmax:float, column:str, weight:Optional[str]=None,
                    title:Optional[str]=None, pyobject:bool=False, draw:bool=False, draw_option:Optional[str]=None):
        self.validate()
        if title is None:
            title = name
        if weight is None:
            r_th1_ptr = self.rdf.Histo1D((name, title, n_bins, xmin, xmax), column)
        else:
            r_th1_ptr = self.rdf.Histo1D((name, title, n_bins, xmin, xmax), column, weight)
        r_th1 = r_th1_ptr.GetPtr()
        if draw:
            if draw_option is None:
                draw_option = ""
            self.create_canvas()
            r_th1.Draw(draw_option)
            self.add_cache(r_th1)
            self.canvas_draw()
        if pyobject:
            py_th1 = TH1(r_th1)
            return py_th1
        return r_th1
    
    def append(self, robject:ROOT.TObject):
        if not isinstance(robject, ROOT.TObject):
            raise RuntimeError("only TObject can be added to TFile")
        self.stdout.info(f"INFO: Added object \"{robject.GetName()}\" to internal components")
        self._components.append(robject)
    
    def add_cache(self, robject:ROOT.TObject):
        self.stdout.info(f"INFO: Added object \"{robject.GetName()}\" to cache")
        self._cache.append(robject)
        
    def clear_cache(self):
        self._cache = []
    
    def save_components(self, fname:str, mode:str="RECREATE", components:Optional[List]=None):
        if components is None:
            components = self.components
        f = ROOT.TFile(fname, mode)
        f.cd()
        for component in components:
            component.Write()
        f.Close()


from typing import Dict, Union, List, Optional

import numpy as np

from quickstats import semistaticmethod

class RooAbsData:
    @staticmethod
    def create_histogram(data:"ROOT.RooAbsData", *args):
        h = data.createHistogram(*args)
        import ROOT
        if isinstance(h, ROOT.TH1):
            from quickstats.interface.root import TH1
            py_h = TH1(h)
        elif isinstance(h, ROOT.TH2):
            from quickstats.interface.root import TH2
            py_h = TH2(h)
        else:
            raise RuntimeError(f"unsupported histogram type: {type(h)}")
        h.Delete()
        return py_h


import os
import click

from .core import DelimitedStr

@click.command(name='build_xml_ws')
@click.option('-i', '--filename', 'source', required=True, 
              help='Input xml file.')
@click.option('--binned/--unbinned', 'use_binned', default=False, show_default=True,
              help='Fit to binned data.')
@click.option('--data_storage_type', default="vector", show_default=True, 
              type=click.Choice(['vector', 'tree', 'composite']),
              help='Set RooAbsData StorageType. Available choices: "vector", "tree", "composite".')
@click.option('-d', '--basedir', default=None, 
              help='Base directory to which files in the xmls are referenced. '
                   'By default, the directory of the input xml file is used.')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def build_xml_ws(**kwargs):
    """
    Build workspace from XML config files
    """
    _kwargs = {}
    for arg_name in ["source", "use_binned", "data_storage_type",
                     "basedir", "verbosity"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['minimizer_config'] = kwargs
    from quickstats.components.workspaces import XMLWSBuilder
    builder = XMLWSBuilder(**_kwargs)
    builder.generate_workspace()
    
@click.command(name='modify_ws')
@click.option('-i', '--filename', 'source', required=True, 
              help='Input xml/json file.')
@click.option('--input_workspace', 
              help='Override input workspace path from the xml/json file.')
@click.option('--output_workspace', 
              help='Override output workspace path from the xml/json file.')
@click.option('--import-class-code/--no-import-class-code', 'import_class_code',
              default=True, show_default=True,
              help='Import class code')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def modify_ws(**kwargs):
    """
    Modify workspace from XML/json config files
    """
    infile = kwargs.pop("input_workspace")
    outfile = kwargs.pop("output_workspace")
    import_class_code = kwargs.pop('import_class_code')
    _kwargs = {}
    for arg_name in ["source", "verbosity"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['minimizer_config'] = kwargs
    from quickstats.components.workspaces import XMLWSModifier
    modifier = XMLWSModifier(**_kwargs)
    modifier.create_modified_workspace(infile=infile, outfile=outfile,
                                       import_class_code=import_class_code)
    
kItemChoices = ['workspace', 'dataset', 'category', 'snapshot', 'pdf',
                'function', 'poi', 'nuisance_parameter', 'global_observable',
                'constrained_nuisance_parameter', 'unconstrained_nuisance_parameter',
                'auxiliary']
kDefaultItems = ",".join(['workspace', 'dataset', 'category', 'poi', 'pdf',
                          'function', 'nuisance_parameter', 'global_observable',
                          'auxiliary'])
kDefaultVisibility = ",".join(["workspace=0b11001", "category=0b10011", "snapshot=0b10011",
                               "dataset=0b11011", "pdf=0b01011", "function=0b01011" , "poi=0b01011",
                               "nuisance_parameter=0b01011", "global_observable=0b01011", "auxiliary=0b01011"])

@click.command(name='compare_ws')
@click.option('-l', '--left', required=True, 
              help='Path to the input workspace file (left of comparison).')
@click.option('-r', '--right', required=True, 
              help='Path to the input workspace file (right of comparison).')
@click.option('--items', cls=DelimitedStr, type=click.Choice(kItemChoices),
              default=kDefaultItems, show_default=True,
              help='Items to include in the comparison (seperated by commas).')
@click.option('--indent', default="   ", show_default=True,
              help='Indentation for each row.')
@click.option('--visibility', default=kDefaultVisibility, show_default=True,
              help='\b\n Set visibility of the items included in the comparison:'
                   '\b\n boolean mask for showing definitions of certain objects'
                   '\b\n 0b00001 = show definitions for unique objects'
                   '\b\n 0b00010 = show definitions for redefined objects'
                   '\b\n 0b00100 = show definitions for reconstituted objects'
                   '\b\n 0b01000 = show definitions for renamed/remapped objects'
                   '\b\n 0b10000 = show definitions for identical objects')
@click.option('--save_text', default=None,
              help='Save summary print out as a text file.')
@click.option('--save_json_data', default=None,
              help='Save comparison data as a json file.')
@click.option('--save_excel_data', default=None,
              help='Save comparison data as an excel file.')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def compare_ws(**kwargs):
    """
    Compare two workspace files
    """
    from quickstats.components.workspaces import WSComparer
    comparer = WSComparer(kwargs["left"], kwargs["right"], items=kwargs["items"],
                          visibility_map=kwargs["visibility"])
    comparer.load_data()
    comparer.process_data()
    summary_str = comparer.get_summary_str(indent=kwargs["indent"])
    print(summary_str)
    save_text = kwargs["save_text"]
    save_json_data = kwargs["save_json_data"]
    save_excel_data = kwargs["save_excel_data"]
    if save_text is not None:
        with open(save_text, "w") as f:
            f.write(summary_str)
    if save_json_data is not None:
        comparer.save_json(save_json_data)
    if save_excel_data is not None:
        comparer.save_excel(save_excel_data)
        
        
@click.command(name='combine_ws')
@click.option('-i', '--filename', 'source', required=True, 
              help='Input xml file.')
@click.option('--input_workspace', 
              help='Override input workspace paths from the xml file. Format: <channel>=<path>,...')
@click.option('--output_workspace', 
              help='Override output workspace path from the xml file.')
@click.option('--save_rename_ws/--skip_rename_ws', default=False, show_default=True,
              help='Save a temporary workspace after the rename step')
@click.option('--save_combine_ws/--skip_combine_ws', default=False, show_default=True,
              help='Save a temporary workspace after the combine step')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('--import-class-code/--no-import-class-code', 'import_class_code',
              default=True, show_default=True,
              help='Import class code')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def combine_ws(**kwargs):
    """
    Combine workspace from XML config files
    """
    infile = kwargs.pop("input_workspace")
    outfile = kwargs.pop("output_workspace")
    import_class_code = kwargs.pop('import_class_code')
    if infile is not None:
        infiles = {}
        for item in infile.split(","):
            subitems = [i.strip() for i in item.split("=")]
            if len(subitems) != 2:
                raise ValueError("invalid format for the argument \"input_workspace\"")
            infiles[subitems[0]] = subitems[1]
    else:
        infiles = None
    save_rename_ws = kwargs.pop("save_rename_ws")
    save_combine_ws = kwargs.pop("save_combine_ws")
    _kwargs = {}
    for arg_name in ["source", "verbosity"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['minimizer_config'] = kwargs
    from quickstats.components.workspaces import XMLWSCombiner
    combiner = XMLWSCombiner(**_kwargs)
    combiner.create_combined_workspace(infiles=infiles, outfile=outfile,
                                       save_rename_ws=save_rename_ws,
                                       save_combine_ws=save_combine_ws,
                                       save_final_ws=True,
                                       import_class_code=import_class_code)    
    


from typing import Optional, Dict, List, Union

import numpy as np

from quickstats.interface.cppyy.vectorize import list2vec, as_np_array
from quickstats import _PRINT_

def unfold_constraints(constraint_pdfs:"ROOT.RooArgSet", observables:"ROOT.RooArgSet",
                       nuisance_parameters:"ROOT.RooArgSet", constraint_cls:Optional[List[str]]=None,
                       recursion_limit:int=50, strip_disconnected:bool=False):
    import ROOT
    result = ROOT.RooArgSet()
    if constraint_cls is None:
        constraint_cls = ROOT.RooFitExt.kConstrPdfClsList
    else:
        
        constraint_cls = list2vec(constraint_cls)
    ROOT.RooFitExt.unfoldConstraints(constraint_pdfs, observables, nuisance_parameters, result,
                                     constraint_cls, 0, recursion_limit, strip_disconnected)
    return result


def convert_argsets(*argsets, fmt:str="argset", to_str:bool=False):
    if to_str and (fmt == "argset"):
        raise ValueError("argset format does not support to_str")
    if fmt == "argset":
        return argsets
    elif fmt in ["list", "dict"]:
        results = []
        if to_str:
            for argset in argsets:
                results.append([i.GetName() for i in argset])
        else:
            for argset in argsets:
                results.append([i for i in argset])
        if fmt == "list":
            return tuple(results)
        elif fmt == "dict":
            dict_result = {}
            for argset, result in zip(argsets, results):
                dict_result[argset.GetName()] = result
            return dict_result
    elif fmt == "series":
        if to_str:
            result = [tuple([i.GetName() for i in items]) for items in zip(*argsets)]
        else:
            result = [tuple(items) for items in zip(*argsets)]
        return result
    else:
        raise ValueError(f"format '{fmt}' not supported")    
            
        
def pair_constraints(constraint_pdfs:"ROOT.RooArgSet", nuisance_parameters:"ROOT.RooArgSet",
                     global_observables:"ROOT.RooArgSet", fmt:str="list", to_str:bool=False, sort:bool=False):
    if sort:
        import ROOT
        ROOT.RooArgSet.sort(constraint_pdfs)
    paired_constraints = ROOT.RooFitExt.pairConstraints(constraint_pdfs, nuisance_parameters, global_observables)
    paired_pdfs  = paired_constraints.pdfs
    paired_nuis  = paired_constraints.nuis
    paired_globs = paired_constraints.globs
    paired_pdfs.setName("pdf")
    paired_nuis.setName("nuis")
    paired_globs.setName("globs")
    return convert_argsets(paired_pdfs, paired_nuis, paired_globs, fmt=fmt, to_str=to_str)

def pair_constraints_base_component(constraint_pdfs:"ROOT.RooArgSet", nuisance_parameters:"ROOT.RooArgSet",
                                    global_observables:"ROOT.RooArgSet", fmt:str="list", to_str:bool=False,
                                    sort:bool=False):
    import ROOT
    if sort:
        ROOT.RooArgSet.sort(constraint_pdfs)
    paired_pdfs      = ROOT.RooArgSet("pdf")
    paired_component = ROOT.RooArgSet("base_component")
    paired_globs     = ROOT.RooArgSet("globs")
    
    for pdf in constraint_pdfs:
        target_component   = None
        target_glob        = None
        pdf_name = pdf.GetName()
        # getting base pdf component
        components = pdf.getComponents()
        components.remove(pdf)
        if components.getSize():
            for c1 in components:
                for c2 in components:
                    if c1 == c2:
                        continue
                    if c2.dependsOn(c1):
                        components.remove(c1)
            if (components.getSize() > 1):
                raise RuntimeError("failed to isolate proper pdf base component")
            elif (components.getSize() == 1):
                target_component = components.first()
        else:
            for nuis in nuisance_parameters:
                if pdf.dependsOn(nuis):
                    target_component = nuis
                    break
        if not target_component:
            _PRINT_.warning(f'WARNING: Could not find base component for the constraint: {pdf_name}')
            continue
        for glob in global_observables:
            if pdf.dependsOn(glob):
                target_glob = glob
                break
        if not target_glob:
            _PRINT_.warning(f'WARNING: Could not find global observable for the constraint: {pdf_name}')
            continue
        paired_pdfs.add(pdf)
        paired_component.add(target_component)
        paired_globs.add(target_glob)
        
    return convert_argsets(paired_pdfs, paired_component, paired_globs, fmt=fmt, to_str=to_str)      
        
def get_str_data(components:"ROOT.RooArgSet", fill_classes:bool=False,
                 fill_definitions:bool=True, content:int=-1,
                 style:int=-1, indent:str="", fmt:str="dict",
                 correction:bool=True):
    import ROOT
    str_data = ROOT.RooFitExt.getStrData(components, fill_classes, fill_definitions,
                                         content, style, indent, correction)
    result = {
        "name" : as_np_array(str_data.names),
    }
    
    if fill_classes:
        result["class"] = as_np_array(str_data.classes)
        
    if fill_definitions:
        result["definition"] = as_np_array(str_data.definitions)
        
    if fmt == "dict":
        return result
    elif fmt == "dataframe":
        import pandas as pd
        return pd.DataFrame(result)
    else:
        raise ValueError(f"format '{fmt}' not supported")


from typing import Optional, Dict, List
import uuid

import ROOT

from .tree_data_source import TreeDataSource

def template_Gaussian(reference_data:TreeDataSource):
    hist = reference_data.create_histogram()
    # evaluate shape properties
    hist_max             = hist.GetMaximum()
    hist_bin_pos_max     = hist.GetMaximumBin()
    hist_pos_max         = hist.GetBinCenter(hist_bin_pos_max)
    hist_pos_FWHM_low    = hist.GetBinCenter(hist.FindFirstBinAbove(0.5 * hist_max))
    hist_pos_FWHM_high   = hist.GetBinCenter(hist.FindLastBinAbove(0.5 * hist_max))
    hist_sigma_effective = (hist_pos_FWHM_high - hist_pos_FWHM_low) / 2.355;
    # free memory
    hist.Delete()
    parameters = {
        'mean'  : ROOT.RooRealVar("mean", "mean", hist_pos_max, hist_pos_FWHM_low, hist_pos_FWHM_high),
        'sigma' : ROOT.RooRealVar("sigma", "sigma", hist_sigma_effective, 0., 5 * hist_sigma_effective)
    }
    return parameters

def template_DSCB(reference_data:TreeDataSource):
    hist = reference_data.create_histogram()
    # evaluate shape properties
    hist_max             = hist.GetMaximum()
    hist_bin_pos_max     = hist.GetMaximumBin()
    hist_pos_max         = hist.GetBinCenter(hist_bin_pos_max)
    hist_pos_FWHM_low    = hist.GetBinCenter(hist.FindFirstBinAbove(0.5 * hist_max))
    hist_pos_FWHM_high   = hist.GetBinCenter(hist.FindLastBinAbove(0.5 * hist_max))
    hist_sigma_effective = (hist_pos_FWHM_high - hist_pos_FWHM_low) / 2.355;
    # free memory
    hist.Delete()
    parameters = {
        'muCBNom'    : ROOT.RooRealVar("muCBNom", "mean of CB", hist_pos_max, hist_pos_FWHM_low, hist_pos_FWHM_high),
        'sigmaCBNom' : ROOT.RooRealVar("sigmaCBNom", "sigma of CB", hist_sigma_effective, 0., 5 * hist_sigma_effective),
        'alphaCBLo'  : ROOT.RooRealVar("alphaCBLo", "Location of transition to a power law on the left", 1, 0., 5.),
        'nCBLo'      : ROOT.RooRealVar("nCBLo", "Exponent of power-law tail on the left", 10, 0., 200.),
        'alphaCBHi'  : ROOT.RooRealVar("alphaCBHi", "Location of transition to a power law on the right", 1, 0., 5.),
        'nCBHi'      : ROOT.RooRealVar("nCBHi", "Exponent of power-law tail on the right", 10, 0., 200.)
    }
    return parameters

def template_Bukin(reference_data:TreeDataSource):
    hist = reference_data.create_histogram()
    # evaluate shape properties
    hist_max             = hist.GetMaximum()
    hist_bin_pos_max     = hist.GetMaximumBin()
    hist_pos_max         = hist.GetBinCenter(hist_bin_pos_max)
    hist_pos_FWHM_low    = hist.GetBinCenter(hist.FindFirstBinAbove(0.5 * hist_max))
    hist_pos_FWHM_high   = hist.GetBinCenter(hist.FindLastBinAbove(0.5 * hist_max))
    hist_sigma_effective = (hist_pos_FWHM_high - hist_pos_FWHM_low) / 2.355;    
    parameters = {
        "Xp"  : ROOT.RooRealVar("Bukin_Xp", "peak position", hist_pos_max, hist_pos_FWHM_low, hist_pos_FWHM_high),
        "sigp": ROOT.RooRealVar("Bukin_sigp", "peak width as FWHM divided by 2*sqrt(2*log(2))=2.35",
                                      hist_sigma_effective, 0.1, 5 * hist_sigma_effective),
        "xi"  : ROOT.RooRealVar("Bukin_xi", "peak asymmetry", 0.0, -1, 1),
        "rho1": ROOT.RooRealVar("Bukin_rho1", "left tail", -0.1, -1.0, 0.0),
        "rho2": ROOT.RooRealVar("Bukin_rho2", "right tail", 0.0, 0.0, 1.0)
    }
    return parameters

def template_ExpGaussExp(reference_data:TreeDataSource):
    parameters = {
        "EGE_mean"  : ROOT.RooRealVar("EGE_mean", "mean of EGE", 124.5, 123., 126.),
        "EGE_sigma" : ROOT.RooRealVar("EGE_sigma", "sigma of EGE", 2.5, 0.5, 8.0),
        "EGE_kLo"   : ROOT.RooRealVar("EGE_kLo", "kLow of EGE", 2.5, 0.01, 10.0),
        "EGE_kHi"   : ROOT.RooRealVar("EGE_kHi", "kHigh of EGE", 2.4, 0.01, 10.0)
    }
    return parameters

def template_Exp(reference_data:TreeDataSource):
    parameters = {
        "Exp_c": ROOT.RooRealVar("Exp_c", "Exp_c", 1, -10, 10)
    }
    return parameters

template_maps = {
    'DSCB': template_DSCB,
    'RooCrystalBall_DSCB': template_DSCB,
    'RooTwoSidedCBShape': template_DSCB,
    'ExpGaussExp': template_ExpGaussExp,
    'RooExpGaussExpShape': template_ExpGaussExp,
    'Bukin': template_Bukin,
    'RooBukinPdf': template_Bukin,
    'Exp': template_Exp,
    'RooExponential': template_Exp,
    'RooGaussian': template_Gaussian,
    'Gaussian': template_Gaussian
}

def get_param_templates(model:str):
    if model not in template_maps:
        raise RuntimeError(f"no default parameter templates found for the model \"{model}\"")
    return template_maps[model]


from .xml_ws_base import XMLWSBase, WSObjectType, XMLAttribType, BracketType
from .sample import Sample
from .systematic import Systematic, SystematicType
from .systematics_domain import SystematicsDomain
from .core_argument_sets import CoreArgumentSets
from .asimov_handler import AsimovHandler
from .xml_ws_builder_v2 import XMLWSBuilder
from .xml_ws_modifier import XMLWSModifier
from .xml_ws_combiner import XMLWSCombiner
from .ws_comparer import WSComparer, ComparisonData, WSItemType


from typing import Dict, Union, List, Optional

import numpy as np

from quickstats import semistaticmethod
from quickstats.interface.cppyy.vectorize import as_np_array

class RooDataSet:
    
    @staticmethod
    def to_numpy(dataset:"ROOT.RooDataSet"):
        import ROOT
        components = dataset.get()
        observables = ROOT.RooArgSet()
        category = None
        for component in components:
            if component.ClassName() == "RooCategory":
                if category is not None:
                    raise RuntimeError("found multiple RooCategory instances")
                category = component
            else:
                observables.add(component)
        if category is None:
            result = ROOT.RFUtils.ExtractData(dataset, observables)
        else:
            result = ROOT.RFUtils.ExtractCategoryData(dataset, observables, category)
        obs_values = as_np_array(result.observable_values)
        n_entries = dataset.numEntries()
        data = {}
        for i, obs in enumerate(observables):
            obs_name = obs.GetName()
            data[obs_name] = obs_values[i*n_entries: (i + 1)*n_entries]
        data['weight'] = as_np_array(result.weights)
        if category is not None:
            data['label'] = as_np_array(result.category_labels)
            data['index'] = as_np_array(result.category_index)
        return data
    
    @staticmethod
    def from_numpy(data:Dict[str, "numpy.ndarray"],
                   variables:"ROOT.RooArgSet",
                   name:str=None, title:str=None,
                   weight_name:str=None, clip_to_limits=True):
        import ROOT
        if name is None:
            name = ""
        if title is None:
            title = ""
            
        if weight_name is None:
            dataset = ROOT.RooDataSet(name, title, variables)
        else:
            dataset = ROOT.RooDataSet(name, title, variables, weight_name)
            
        real_variables = {}
        cat_variables = {}
        for v in variables:
            var_name = v.GetName()
            class_name = v.ClassName()
            if var_name == weight_name:
                continue
            if var_name not in data:
                if (class_name == "RooCategory") and ("index" in data):
                    var_name = "index"
                else:
                    raise RuntimeError(f"missing data for the variable `{var_name}`")
            if class_name == "RooCategory":
                cat_variables[var_name] = v
            else:
                real_variables[var_name] = v
                if clip_to_limits:
                    data[var_name] = np.clip(data[var_name], 
                                             a_min=v.getMin(),
                                             a_max=v.getMax())
        if weight_name not in data:
            if "weight" in data:
                weight_name = "weight"
            else:
                raise RuntimeError(f"missing data for the variable `{weight_name}`")
        data_sizes = [len(data[k]) for k in data]
        if len(set(data_sizes)) > 1:
            raise RuntimeError("data has inconsistent sizes")
        if weight_name is None:
            for i in range(data_sizes[0]):
                for name, variable in real_variables.items():
                    variable.setVal(data[name][i])
                for name, variable in cat_variables.items():
                    variable.setIndex(data[name][i])
                dataset.add(variables, 1.)
        else:
            for i in range(data_sizes[0]):
                for name, variable in real_variables.items():
                    variable.setVal(data[name][i])
                for name, variable in cat_variables.items():
                    variable.setIndex(int(data[name][i]))
                weight = data[weight_name][i]
                dataset.add(variables, weight)
        return dataset
    
    @semistaticmethod
    def to_pandas(self, dataset:"ROOT.RooDataSet"):
        numpy_data = self.to_numpy(dataset)
        import pandas as pd
        df = pd.DataFrame(numpy_data)
        return df
    
    @staticmethod
    def get_dataset_map(dataset_dict:Dict):
        import ROOT
        dsmap = ROOT.std.map('string, RooDataSet*')()
        dsmap.keepalive = list()
        for c, d in dataset_dict.items():
            dsmap.keepalive.append(d)
            dsmap.insert(dsmap.begin(), ROOT.std.pair("const string, RooDataSet*")(c, d))
        return dsmap
    
    @staticmethod
    def from_RooDataHist(source:"ROOT.RooDataHist", pdf:"ROOT.RooAbsPdf",
                         name:Optional[str]=None):
        import ROOT
        if name is None:
            name = source.GetName()
        parameters = source.get()
        category = None
        for parameter in parameters:
            if parameter.ClassName() == "RooCategory":
                category = parameter
                break
      # case multi-category data
        if category is not None:
            dataset_map = {}
            data_cat = source.split(category, True)
            n_cat = len(category)
            observables = ROOT.RooArgSet()
            for i in range(n_cat):
                category.setBin(i)
                cat_name = category.getLabel()
                pdf_i = pdf.getPdf(cat_name)
                data_i = data_cat.FindObject(cat_name)                
                obs_i = pdf_i.getObservables(data_i).first()
                _obs_i = data_i.get().find(obs_i.GetName())
                w_i = ROOT.RooRealVar(f"weight_{i}", f"weight_{i}", 1)
                dataset_i = ROOT.RooDataSet(f"dataset_{i}", f"dataset_{i}",
                                            ROOT.RooArgSet(obs_i, w_i),
                                            ROOT.RooFit.WeightVar(w_i))
                ROOT.RFUtils.CopyData(data_i, dataset_i, _obs_i, obs_i, w_i)
                dataset_map[cat_name] = dataset_i
                observables.add(obs_i)
            w = ROOT.RooRealVar("weight", "weight", 1)
            observables.add(w)
            cpp_dataset_map = RooDataSet.get_dataset_map(dataset_map)
            dataset = ROOT.RooDataSet(name, name, observables,
                                      ROOT.RooFit.Index(category),
                                      ROOT.RooFit.Import(cpp_dataset_map),
                                      ROOT.RooFit.WeightVar(w))
        # case single-category data
        else:
            obs = pdf.getObservables(source).first()
            _obs = source.get().find(obs.GetName())
            w = ROOT.RooRealVar("weight", "weight", 1)
            dataset = ROOT.RooDataSet(name, name, ROOT.RooArgSet(obs, w),
                                      ROOT.RooFit.WeightVar(w))
            ROOT.RFUtils.CopyData(source, dataset, _obs, obs, w)
        return dataset

    @staticmethod
    def _get_cat_and_obs(variables:"ROOT.RooArgSet"):
        cat_variable = None
        observables = {}
        for v in variables:
            class_name = v.ClassName()
            if class_name == "RooCategory":
                if cat_variable is not None:
                    raise RuntimeError("found multiple RooCategory instances")
                cat_variable = v
            else:
                var_name = v.GetName()
                observables[var_name] = v
        if cat_variable is None:
            raise RuntimeError("missing RooCategory instance from variables")
        return cat_variable, observables

    @semistaticmethod
    def create_binned_category_dataset(self, data:Dict[str, "numpy.ndarray"],
                                       pdf:"ROOT.RooAbsPdf",
                                       variables:"ROOT.RooArgSet",
                                       weight_name:str="weightVar",
                                       name:str=None, title:str=None):
        import ROOT
        if name is None:
            name = ""
        if title is None:
            title = ""        
        cat_variable, observables = self._get_cat_and_obs(variables)
        n_cat = cat_variable.size()
        cat_names = []
        cat_obs_names = []
        for i in range(n_cat):
            cat_variable.setIndex(i)
            cat_name = cat_variable.getLabel()
            cat_names.append(cat_name)
            pdf_cat = pdf.getPdf(cat_name)
            obs = pdf_cat.getObservables(variables)
            cat_obs = obs.first()
            cat_obs_names.append(cat_obs.GetName())
        if set(cat_obs_names) != set(observables):
            raise RuntimeError("the given variables are insistent with the category observables")
        if not set(cat_names).issubset(set(data)):
            missing = list(set(cat_names) - set(data))
            raise RuntimeError("missing data for the following categories: {}".format(",".join(missing)))
        dataset = ROOT.RooDataSet(name, title, variables, weight_name)
        for i, (cat_name, obs_name) in enumerate(zip(cat_names, cat_obs_names)):
            observable = observables[obs_name]
            data_i = data[cat_name]
            cat_variable.setIndex(i)
            n_bins = observable.getBins()
            n_bins_data = len(data_i)
            if n_bins_data != n_bins:
                raise RuntimeError(f"the observable has `{n_bins}` bins but data has `{n_bins_data}`")
            for j in range(n_bins_data):
                observable.setBin(j)
                dataset.add(variables, data_i[j])
        return dataset
    
    @semistaticmethod
    def to_category_data(self, dataset:"ROOT.RooDataSet",
                         pdf:"ROOT.RooAbsPdf"):
        variables = dataset.get()
        cat_variable, observables = self._get_cat_and_obs(variables)
        n_cat = cat_variable.size()
        data = self.to_numpy(dataset)
        result = {}
        for i in range(n_cat):
            cat_variable.setIndex(i)
            cat_name = cat_variable.getLabel()
            pdf_cat = pdf.getPdf(cat_name)
            obs = pdf_cat.getObservables(variables)
            cat_obs = obs.first()
            obs_name = cat_obs.GetName()
            mask = (data['index'] == i)
            bin_value = data[obs_name][mask]
            ind = np.argsort(bin_value)
            bin_weight = data['weight'][mask][ind]
            result[cat_name] = bin_weight
        return result
    
    @staticmethod
    def fill_from_TH1(dataset:"ROOT.RooDataSet", hist:"ROOT.TH1",
                      skip_out_of_range:bool=True,
                      blind_range:Optional[List[float]]=None,
                      min_bin_value:float=0,
                      weight_scale:float=1):
        import ROOT
        parameters = dataset.get()
        if parameters.size() > 1:
            raise RuntimeError("multiple observables are not allowed")
        x = parameters.first()
        weight_var = dataset.weightVar()
        # blinding will be taken care of
        xmin = x.getMin()
        xmax = x.getMax()
        n_bins = hist.GetNbinsX()
        for i in range(1, n_bins + 1):
            bin_center = hist.GetBinCenter(i)
            # skip bins that are out of range
            if skip_out_of_range and ((bin_center > xmax) or (bin_center < xmin)):
                continue
            # skip bins in the blind range
            if (blind_range and (bin_center > blind_range[0]) and (bin_center < blind_range[1])):
                continue
            x.setVal(bin_center)
            bin_content = hist.GetBinContent(i)
            weight = bin_content * weight_scale
            # if the weight is negligible, consider it as zero
            if (weight < min_bin_value):
                continue;
            if weight_var:
                weight_var.setVal(weight)
                dataset.add(ROOT.RooArgSet(x, weight_var), weight)
            else:
                dataset.add(ROOT.RooArgSet(x), weight)
    
    @staticmethod
    def get_x_and_weight(dataset:"ROOT.RooDataSet"):
        parameters = dataset.get()
        if parameters.size() > 1:
            raise RuntimeError("multiple observables are not allowed")
        x = parameters.first()
        weight_var = dataset.weightVar()
        return x, weight_var
    
    @staticmethod
    def to_TH1(dataset:"ROOT.RooDataSet", name:str,
               blind_range:Optional[List[float]]=None,
               weight_scale:float=1):
        x, weight_var = RooDataSet.get_x_and_weight(dataset)
        n_bins = x.getBins()
        x_min = x.getMin()
        x_max = x.getMax()
        import ROOT
        hist = ROOT.TH1D(name, name, n_bins, x_min, x_max)
        hist.Sumw2()
        for i in range(dataset.numEntries()):
            dataset.get(i)
            x_val = x.getVal()
            obs.setVal(x_val)
            weight = dataset.weight() * weight_scale
            # ignore data in the blind range
            if (blind_range and (x_val > blind_range[0]) and (x_val < blind_range[1])):
                continue
            hist.Fill(x_val, weight)
        return hist
    
    @staticmethod
    def add_ghost_weights(dataset:"ROOT.RooDataSet",
                          blind_range:Optional[List[float]]=None,
                          ghost_weight:float=1e-9):
        x, weight_var = RooDataSet.get_x_and_weight(dataset)
        xmin = x.getMin()
        xmax = x.getMax()
        n_bins = x.getBins()
        bin_width = (xmax - xmin) / n_bins
        data = RooDataSet.to_numpy(dataset)
        x_data = data[x.GetName()]
        weight_data = data["weight"]
        hist, bin_edges = np.histogram(x_data, bins=n_bins, range=(xmin, xmax), weights=weight_data)
        from quickstats.maths.statistics import bin_edge_to_bin_center
        bin_centers = bin_edge_to_bin_center(bin_edges)
        import ROOT
        for bin_val, bin_center in zip(hist, bin_centers):
            if (bin_val != 0):
                continue
            if (blind_range and (bin_center > blind_range[0]) and (bin_center < blind_range[1])):
                continue
            x.setVal(bin_center)
            weight_var.setVal(ghost_weight)
            dataset.add(ROOT.RooArgSet(x, weight_var), ghost_weight)


from typing import Optional, Union, List, Dict
import numpy as np

class TFile:
    
    def __init__(self, source:Union[str, "ROOT.TFile"]):
        self.initializes(source)
        
    def initialize(self, source:Union[str, "ROOT.TFile"]):
        self.rfile = TFile._open(source)
        
    @staticmethod
    def is_corrupt(f:"ROOT.TFile"):
        if f.IsZombie():
            return True
        if f.GetNkeys() == 0:
            return True
        import ROOT
        if f.TestBit(ROOT.TFile.kRecovered):
            return True
        return False
    
    @staticmethod
    def _open(source:Union[str, "ROOT.TFile"]):
        # source is path to a root file
        if isinstance(source, str):
            import ROOT
            source = ROOT.TFile(source)
            
        if TFile.is_corrupt(source):
            raise RuntimeError(f'root file "{source.GetName()}" is corrupt')
            
        return source        
        
    """
    def make_branches(self, branch_data):
        branches = {}
        return branches
    
    def fill_branches(self, treename:str, branch_data):
        if self.obj is None:
            raise RuntimeError("no active ROOT file instance defined")
        tree = self.obj.Get(treename)
        if not tree:
            raise RuntimeError(f"the ROOT file does not contain the tree named \"{treename}\"")
        n_entries = tree.GetEntriesFast()
        
        for i in range(n_entries):
            for branch in branches:
                
        tree.SetDirectory(self.obj)
        # save only the new version of the tree
        tree.GetCurrentFile().Write("", ROOT.TObject.kOverwrite)
    """

    def close(self):
        self.rfile.Close()
        self.rfile = None


from typing import Union, Optional, List, Dict, Tuple

import math
import numpy as np

def calculate_nll(obs:float, exp:float):
    import ROOT
    return np.log(ROOT.TMath.Poisson(obs, exp))

def calculate_chi2(data_obs, data_exp, error_obs=None, threshold:float=3, epsilon:float=1e-6):
    if np.any(data_obs < 0):
        raise RuntimeError("data observed has negative-value element(s)")
    if np.any(data_exp < 0):
        raise RuntimeError("data expected has negative-value element(s)")        
    if error_obs is None:
        error_obs = np.sqrt(data_obs)
    data_obs = np.array(data_obs, dtype=np.float64)
    data_exp = np.array(data_exp, dtype=np.float64)
    error_obs = np.array(error_obs, dtype=np.float64)
    if data_obs.shape != data_exp.shape:
        raise RuntimeError("data observed and data expected have different shapes")
    if data_obs.shape != error_obs.shape:
        raise RuntimeError("data observed and error observed have different shapes")
    if data_obs.ndim != 1:
        raise RuntimeError("only one dimensional data is supported")
    chi2, chi2_last, obs_aggregate, exp_aggregate, error2_aggregate = 0., 0., 0., 0., 0.
    nbin_chi2 = 0
    bin_last = 1
    n_bins = len(data_obs)
    for i in range(n_bins):
        obs_aggregate += data_obs[i]
        exp_aggregate += data_exp[i]
        error2_aggregate += error_obs[i] ** 2
        if (obs_aggregate / np.sqrt(error2_aggregate) < threshold) or \
           (abs(obs_aggregate) < epsilon):
            if i != (n_bins - 1):
                continue
            else:
                chi2 -= chi2_last
                obs_aggregate = np.sum(data_obs[bin_last:])
                exp_aggregate = np.sum(data_exp[bin_last:])
                error2_aggregate = np.sum(error_obs[bin_last:] ** 2)
                chi2 += ((obs_aggregate - exp_aggregate) / np.sqrt(error2_aggregate)) ** 2
                if nbin_chi2 == 0:
                    nbin_chi2 += 1
        else:
            chi2_last = ((obs_aggregate - exp_aggregate) / np.sqrt(error2_aggregate)) ** 2
            bin_last = i
            chi2 += chi2_last
            nbin_chi2 += 1
            obs_aggregate, exp_aggregate, error2_aggregate = 0., 0., 0.
    # calculate likelihood
    nll, nll_last, nll_sat, nll_sat_last = 0., 0., 0., 0.
    obs_aggregate, exp_aggregate = 0., 0.
    nbin_nll = 0
    bin_last = 0
    for i in range(n_bins):
        obs_aggregate += data_obs[i]
        exp_aggregate += data_exp[i]
        error2_aggregate += error_obs[i] ** 2
        if (obs_aggregate < 2):
            if i != (n_bins - 1):
                continue
            else:
                nll -= nll_last
                nll_sat -= nll_sat_last
                obs_aggregate = np.sum(data_obs[bin_last:])
                exp_aggregate = np.sum(data_exp[bin_last:])
                nll += -1 * self.calculate_nll(obs_aggregate, exp_aggregate)
                # saturated
                nll_sat += -1 * self.calculate_nll(obs_aggregate, obs_aggregate)
                if nbin_nll == 0:
                    nbin_nll += 1
        else:
            nll_last = -1 * self.calculate_nll(obs_aggregate, exp_aggregate)
            nll_sat_last = -1 * self.calculate_nll(obs_aggregate, obs_aggregate)
            nll += nll_last
            nll_sat += nll_sat_last
            bin_last = i
            nbin_nll += 1
            obs_aggregate, exp_aggregate = 0., 0.
    result = {
        'chi2': chi2,
        'nbin_chi2': nbin_chi2,
        'nll': nll,
        'nll_sat': nll_sat,
        'nbin_nll': nbin_nll
    }
    return result

def get_poisson_interval(data:np.ndarray, n_sigma:float=1):
    """
        Calculate the Poisson error interval for binned data.
        
        Arguments:
            data: np.ndarray
                Array containing the event number in each bin.
            n_simg: float
                Number of sigma to use for the Poisson interval.
    """
    from quickstats.interface.root import TH1
    return TH1.GetPoissonError(data, n_sigma)

def get_counting_significance(s:float, b:float, sigma_b:float=0, leading_order:bool=False):
    """
        Asimov approximation for the median significance in a counting experiment.
        
        Arguments:
            s: float
                Expected number of signal events.
            b: float
                Expected number of background events.
            sigma_b: float, default = 0
                Background uncertainty. A zero value means the number of
                background events is exactly known.
            leading_order: bool, default=False
                Whether to use leading order approximation.
    """
    if sigma_b == 0:
        if leading_order:
            return s / np.sqrt(b)
        n = s + b
        return np.sqrt(2 * ((n * np.log(n / b)) - s))
    else:
        sigma_b2 = sigma_b * sigma_b
        if leading_order:
            return s / np.sqrt(b + sigma_b2)
        n = s + b
        b_plus_sigma2 = b + sigma_b2
        first_term = n * np.log((n * b_plus_sigma2)/(b * b + n * sigma_b2))
        second_term = b * b / sigma_b2 * np.log(1 + (sigma_b2 * s) / (b * b_plus_sigma2))
        return np.sqrt(2 * (first_term - second_term))
    
def get_combined_counting_significance(s:np.ndarray, b:np.ndarray,
                                       sigma_b:Union[np.ndarray, float]=0,
                                       leading_order:bool=False):
    """
        Combined significance in multiple independent counting experiments.
        
        Arguments:
            s: np.ndarray of float
                Array of expected number of signal events in each experiment.
            b: np.ndarray of float
                Array of expected number of background events in each experiment.
            sigma_b: float / np.ndarray of float, default = 0
                Array of background uncertainties in each experiment. A zero value
                means the number of background events is exactly known.
            leading_order: bool, default=False
                Whether to use leading order approximation.
    """
    if sigma_b == 0:
        if leading_order:
            Z2 = s * s / b
        else:
            n = s + b
            Z2 = 2 * ((n * np.log(n / b)) - s)
    else:
        sigma_b2 = sigma_b * sigma_b
        if leading_order:
            Z2 = s * s / (b + sigma_b2)
        else:
            n = s + b
            b_plus_sigma2 = b + sigma_b2
            first_term = n * np.log((n * b_plus_sigma2)/(b * b + n * sigma_b2))
            second_term = b * b / sigma_b2 * np.log(1 + (sigma_b2 * s) / (b * b_plus_sigma2))
            Z2 = 2 * (first_term - second_term)
    if Z2.ndim > 1:
        Z_combined = np.sqrt(np.sum(Z2, axis=Z2.ndim - 1))
    else:
        Z_combined = np.sqrt(np.sum(Z2))
    return Z_combined

def bin_edge_to_bin_center(bin_edge:np.ndarray):
    return 0.5 * (bin_edge[1:] + bin_edge[:-1])

def min_max_to_range(min_val:Optional[float]=None, max_val:Optional[float]=None):
    if (min_val is None) and (max_val is None):
        return None
    if (min_val is not None) and (max_val is not None):
        return (min_val, max_val)
    raise ValueError("min and max values must be all None or all float")

def get_hist_data(x:np.ndarray, weights:Optional[np.ndarray]=None,
                  normalize:bool=True,
                  range:Optional[Union[List, Tuple]]=None,
                  bins:int=10,
                  error_method:str="auto"):
    x = np.array(x)
    if weights is None:
        if normalize:
            weights = np.ones(x.shape) / len(x)
        y, bin_edges = np.histogram(x, bins=bins, range=range, weights=weights)
        bin_centers = bin_edge_to_bin_center(bin_edges)
        hist_data = {
            "x"    : bin_centers,
            "y"    : y,
            "xerr" : None,
            "yerr" : None
        }
    else:
        weights = np.array(weights)
        error_method = error_method.lower()
        assert error_method in ["sumw2", "poisson", "auto"]
        if error_method == "auto":
            unit_weight = np.allclose(weights, np.ones(weights.shape))
            error_method = "poisson" if unit_weight else "sumw2"
        if normalize:
            size = weights.sum()
        else:
            size = 1
        if error_method == "poisson":
            _y, bin_edges = np.histogram(x, bins=bins, range=range, weights=weights)
            from quickstats.maths.statistics import get_poisson_interval
            pois_interval = get_poisson_interval(_y)
            yerr =  (pois_interval["lo"] / size, pois_interval["hi"] / size)
            _weights = weights / size
            y, bin_edges = np.histogram(x, bins=bins, range=range, weights=_weights)
        elif error_method == "sumw2":
            _y, bin_edges = np.histogram(x, bins=bins, range=range, weights=weights**2)
            _weights = weights / size
            y, bin_edges = np.histogram(x, bins=bins, range=range, weights=_weights)
            yerr = np.sqrt(_y) / size
        bin_centers = bin_edge_to_bin_center(bin_edges)
        xerr = (bin_centers[1] - bin_centers[0])/2
        hist_data = {
            "x"    : bin_centers,
            "y"    : y,
            "xerr" : xerr,
            "yerr" : yerr
        }
    return hist_data
    